import json
from dataclasses import dataclass
from typing import Any

from vigorish.util.list_helpers import as_dict_list
from vigorish.util.string_helpers import get_brooks_team_id, validate_bbref_game_id


@dataclass
class BBRefBoxscore:
    """Batting and pitching statistics for a single MLB game."""

    boxscore_url: str
    bbref_game_id: str
    game_meta_info: Any
    away_team_data: Any
    home_team_data: Any
    innings_list: Any
    umpires: Any
    player_id_match_log: Any
    player_team_dict: Any
    player_name_dict: Any

    @property
    def upload_id(self):
        return self.bbref_game_id

    @property
    def bb_game_id(self):
        result = validate_bbref_game_id(self.bbref_game_id)
        game_date = result.value["game_date"]
        game_number = result.value["game_number"]
        away_team_id = get_brooks_team_id(self.away_team_data.team_id_br).lower()
        home_team_id = get_brooks_team_id(self.home_team_data.team_id_br).lower()
        return (
            f"gid_{game_date.year}_{game_date.month:02d}_{game_date.day:02d}_"
            f"{away_team_id}mlb_{home_team_id}mlb_{game_number}"
        )

    @property
    def game_id_dict(self):
        return {f"{self.bbref_game_id}": f"{self.bb_game_id}"}

    @property
    def game_date(self):
        result = validate_bbref_game_id(self.bbref_game_id)
        return result.value["game_date"] if result.success else None

    @property
    def away_team_pitch_appearance_count(self):
        return len(self.away_team_data.pitching_stats)

    @property
    def home_team_pitch_appearance_count(self):
        return len(self.home_team_data.pitching_stats)

    @property
    def pitch_appearance_count(self):
        return self.away_team_pitch_appearance_count + self.home_team_pitch_appearance_count

    @property
    def away_team_pitch_count(self):
        return sum(
            int(pitch_stats.pitch_count) for pitch_stats in self.away_team_data.pitching_stats
        )

    @property
    def home_team_pitch_count(self):
        return sum(
            int(pitch_stats.pitch_count) for pitch_stats in self.home_team_data.pitching_stats
        )

    @property
    def pitch_count(self):
        return self.away_team_pitch_count + self.home_team_pitch_count

    @property
    def pitch_appearances(self):
        mlb_id_name_dict = {v: k for k, v in self.player_name_dict.items()}
        pitch_apps = self.away_team_data.pitching_stats.copy()
        pitch_apps.extend(self.home_team_data.pitching_stats.copy())
        pitch_apps_dicts = [pa.as_dict() for pa in pitch_apps]
        for app_dict in pitch_apps_dicts:
            player_id = app_dict["player_id_br"]
            app_dict["player_name"] = mlb_id_name_dict[player_id]
        return pitch_apps_dicts

    def __repr__(self):
        return f"<BBRefBoxscore bbref_game_id={self.bbref_game_id}>"

    def as_dict(self):
        """Convert boxscore to a dictionary."""
        return dict(
            __bbref_boxscore__=True,
            boxscore_url=self.boxscore_url,
            bbref_game_id=self.bbref_game_id,
            game_meta_info=self.game_meta_info.as_dict(),
            away_team_data=self.away_team_data.as_dict(),
            home_team_data=self.home_team_data.as_dict(),
            innings_list=as_dict_list(self.innings_list),
            player_id_match_log=self.player_id_match_log,
            umpires=as_dict_list(self.umpires),
            player_team_dict=self.player_team_dict,
            player_name_dict=self.player_name_dict,
        )

    def as_json(self):
        """Convert boxscore to JSON."""
        return json.dumps(self.as_dict(), indent=2, sort_keys=False)



class BBRefUmpire:
    """Name and field location of one umpire for a single game."""

    field_location = ""
    umpire_name = ""

    def as_dict(self):
        """Convert umpire name and field location to a dictionary."""
        dict = {
            "field_location": "{loc}".format(loc=self.field_location),
            "umpire_name": "{n}".format(n=self.umpire_name),
        }
        return dict



"""Menu that allows the user to view data for games where all data has been scraped."""
import subprocess

from bullet import Input
from getch import pause

from vigorish.cli.components.prompts import select_game_prompt, user_options_prompt
from vigorish.cli.components.util import print_error, print_heading
from vigorish.cli.menu_item import MenuItem
from vigorish.cli.menu_items.view_game_data import ViewGameData
from vigorish.config.database import GameScrapeStatus, Team
from vigorish.constants import EMOJI_DICT, MENU_NUMBERS
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id


class ViewGameDataMenu(MenuItem):
    def __init__(self, app, audit_report):
        super().__init__(app)
        self.audit_report = audit_report
        self.menu_item_text = "View Scraped Game Data"
        self.menu_item_emoji = EMOJI_DICT.get("MICROSCOPE", "")
        self.exit_menu = False
        self.menu_option = None
        self.mlb_season = None
        self.team_id = None
        self.game_id = None

    def launch(self):
        while True:
            result = self.prompt_user_choose_method_to_select_game()
            if result.failure:
                break
            self.menu_option = result.value
            while True:
                if self.menu_option == "MANUAL":
                    result = self.enter_game_id()
                if self.menu_option == "SEASON":
                    result = self.select_season_game_id()
                if self.menu_option == "TEAM":
                    result = self.select_team_game_id()
                if result.failure:
                    break
        return Result.Ok()

    def prompt_user_choose_method_to_select_game(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Please choose an option below", fg="bright_yellow")
        prompt = "How would you like to locate the game you wish to view?"
        choices = {
            f"{MENU_NUMBERS.get(1)}  Enter Game ID Manually": "MANUAL",
            f"{MENU_NUMBERS.get(2)}  Browse by Team": "TEAM",
            f"{MENU_NUMBERS.get(3)}  Browse by Season": "SEASON",
            f"{EMOJI_DICT.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, prompt, clear_screen=False)

    def enter_game_id(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Enter a BBRef Game ID", fg="bright_yellow")
        check_game_id = Input("Enter a BBRef Game ID: ").launch()
        try:
            result = validate_bbref_game_id(check_game_id)
        except ValueError:
            error = f"\n'{check_game_id}' is NOT a valid BBRef Game ID, please try again."
            print_error(error)
            pause(message="Press any key to continue...")
            return Result.Fail("")
        self.game_id = result.value["game_id"]
        game_date = result.value["game_date"]
        all_valid_game_ids = self.audit_report[game_date.year]["successful"]
        if self.game_id not in all_valid_game_ids:
            error = f"\nRequirements to show data for {check_game_id} have not been met"
            print_error(error)
            pause(message="Press any key to continue...")
            return Result.Fail("")
        view_game_data = ViewGameData(self.app, self.game_id)
        view_game_data.launch()
        return Result.Fail("")

    def select_season_game_id(self):
        while True:
            result = self.select_season_prompt()
            if result.failure:
                return result
            self.mlb_season = result.value
            game_ids = self.audit_report[self.mlb_season]["successful"]
            while True:
                subprocess.run(["clear"])
                heading = f"Scraped Data Viewer - Select Game (MLB Season: {self.mlb_season})"
                print_heading(heading, fg="bright_yellow")
                result = select_game_prompt(game_ids, use_numbers=False, clear_screen=False)
                if result.failure:
                    break
                self.game_id = result.value
                view_game_data = ViewGameData(self.app, self.game_id)
                view_game_data.launch()

    def select_team_game_id(self):
        while True:
            result = self.select_season_prompt()
            if result.failure:
                return result
            self.mlb_season = result.value
            while True:
                result = self.select_team_prompt()
                if result.failure:
                    self.team_id = None
                    break
                game_ids = result.value
                while True:
                    subprocess.run(["clear"])
                    heading = (
                        "Scraped Data Viewer - Select Game "
                        f"(MLB Season: {self.mlb_season}, Team: {self.team_id})"
                    )
                    print_heading(heading, fg="bright_yellow")
                    result = select_game_prompt(game_ids, use_numbers=False, clear_screen=False)
                    if result.failure:
                        break
                    self.game_id = result.value
                    view_game_data = ViewGameData(self.app, self.game_id)
                    view_game_data.launch()

    def select_season_prompt(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Select MLB Season", fg="bright_yellow")
        prompt = "Select an MLB season from the list below:"
        choices = {
            f"{MENU_NUMBERS.get(num, str(num))}  {year}": year
            for num, (year, results) in enumerate(self.audit_report.items(), start=1)
            if results["successful"]
        }
        choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
        return user_options_prompt(choices, prompt, clear_screen=False)

    def select_team_prompt(self):
        team_choices_dict = {
            t.team_id_br: t.name
            for t in Team.get_all_teams_for_season(self.db_session, year=self.mlb_season)
        }
        subprocess.run(["clear"])
        heading = f"Scraped Data Viewer - Select Team (MLB Season: {self.mlb_season})"
        print_heading(heading, fg="bright_yellow")
        prompt = "Select a team from the list below::"
        choices = {
            f"{EMOJI_DICT.get('BLUE_DIAMOND')}  {name} ({team_id})": team_id
            for team_id, name in team_choices_dict.items()
        }
        choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
        result = user_options_prompt(choices, prompt, clear_screen=False)
        if result.failure:
            return result
        self.team_id = result.value
        return self.get_team_game_ids()

    def get_team_game_ids(self):
        all_valid_game_ids = self.audit_report[self.mlb_season]["successful"]
        team_game_ids = GameScrapeStatus.get_all_bbref_game_ids_for_team(
            self.db_session, self.team_id, self.mlb_season
        )
        valid_team_game_ids = list(set(team_game_ids).intersection(set(all_valid_game_ids)))
        game_id_weights = self.get_game_id_weights(valid_team_game_ids)
        valid_team_game_ids.sort(key=lambda x: game_id_weights[x])
        return Result.Ok(valid_team_game_ids)

    def get_game_id_weights(self, game_ids):
        return {g: validate_bbref_game_id(g).value["game_date"] for g in game_ids}



from vigorish.config.database import GameScrapeStatus
from vigorish.enums import DataSet
from vigorish.util.result import Result


def update_data_set_bbref_boxscores(scraped_data, db_session, season):
    result = scraped_data.get_all_scraped_bbref_game_ids(season.year)
    if result.failure:
        return result
    scraped_bbref_gameids = result.value
    unscraped_bbref_gameids = GameScrapeStatus.get_all_unscraped_bbref_game_ids_for_season(
        db_session, season.id
    )
    new_bbref_game_ids = set(scraped_bbref_gameids) & set(unscraped_bbref_gameids)
    if not new_bbref_game_ids:
        return Result.Ok([])
    result = update_status_bbref_boxscore_list(scraped_data, db_session, new_bbref_game_ids)
    if result.failure:
        return result
    return Result.Ok()


def update_status_bbref_boxscore_list(scraped_data, db_session, new_bbref_game_ids):
    for bbref_game_id in new_bbref_game_ids:
        boxscore = scraped_data.get_bbref_boxscore(bbref_game_id)
        if not boxscore:
            error = f"Failed to retrieve {DataSet.BBREF_BOXSCORES} (URL ID: {bbref_game_id})"
            return Result.Fail(error)
        result = update_status_bbref_boxscore(db_session, boxscore)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_status_bbref_boxscore(db_session, boxscore):
    try:
        game_status = GameScrapeStatus.find_by_bbref_game_id(db_session, boxscore.bbref_game_id)
        if not game_status:
            error = (
                f"scrape_status_game does not contain an "
                f"entry for bbref_game_id: {boxscore.bbref_game_id}"
            )
            return Result.Fail(error)
        game_status.scraped_bbref_boxscore = 1
        game_status.pitch_app_count_bbref = boxscore.pitch_appearance_count
        game_status.total_pitch_count_bbref = boxscore.pitch_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")



"""Pitchfx measurements for a single pitch scraped from brooksbaseball.com."""
from dataclasses import dataclass
from datetime import datetime, timezone

from vigorish.util.datetime_util import TIME_ZONE_NEW_YORK
from vigorish.util.dt_format_strings import DT_AWARE
from vigorish.util.regex import PFX_TIMESTAMP_REGEX
from vigorish.util.string_helpers import validate_bbref_game_id


@dataclass
class BrooksPitchFxData:
    """Pitchfx measurements for a single pitch scraped from brooksbaseball.com."""

    pitcher_name: str = ""
    pitch_app_id: str = ""
    pitcher_id: str = "0"
    batter_id: str = "0"
    pitcher_team_id_bb: str = ""
    opponent_team_id_bb: str = ""
    bb_game_id: str = ""
    bbref_game_id: str = ""
    table_row_number: str = "0"
    park_sv_id: str = ""
    play_guid: str = ""
    ab_total: str = "0"
    ab_count: str = "0"
    ab_id: str = "0"
    des: str = ""
    type: str = ""
    id: str = "0"
    sz_top: str = "0"
    sz_bot: str = "0"
    pfx_xdatafile: str = "0"
    pfx_zdatafile: str = "0"
    mlbam_pitch_name: str = ""
    zone_location: str = "0"
    pitch_con: str = "0"
    stand: str = ""
    strikes: str = "0"
    balls: str = "0"
    p_throws: str = ""
    pdes: str = ""
    spin: str = "0"
    norm_ht: str = "0"
    inning: str = "0"
    tstart: str = "0"
    vystart: str = "0"
    ftime: str = "0"
    pfx_x: str = "0"
    pfx_z: str = "0"
    uncorrected_pfx_x: str = "0"
    uncorrected_pfx_z: str = "0"
    x0: str = "0"
    y0: str = "0"
    z0: str = "0"
    vx0: str = "0"
    vy0: str = "0"
    vz0: str = "0"
    ax: str = "0"
    ay: str = "0"
    az: str = "0"
    start_speed: str = "0"
    px: str = "0"
    pz: str = "0"
    pxold: str = "0"
    pzold: str = "0"
    tm_spin: str = "0"
    sb: str = "0"
    game_start_time_str: str = ""
    time_pitch_thrown_str: str = ""
    has_zone_location: bool = False
    is_patched: bool = False
    is_duplicate_guid: bool = False
    is_duplicate_pitch_number: bool = False
    is_invalid_ibb: bool = False
    is_out_of_sequence: bool = False

    @property
    def game_start_time(self):
        return (
            datetime.strptime(self.game_start_time_str, DT_AWARE)
            if self.game_start_time_str
            else None
        )

    @property
    def time_pitch_thrown(self):
        match = PFX_TIMESTAMP_REGEX.match(self.park_sv_id)
        if not match:
            return None
        group_dict = match.groupdict()
        if int(group_dict["second"]) >= 60:
            group_dict["second"] = 0
        game_dict = validate_bbref_game_id(self.bbref_game_id).value
        timestamp = datetime(
            game_dict["game_date"].year,
            int(group_dict["month"]),
            int(group_dict["day"]),
            int(group_dict["hour"]),
            int(group_dict["minute"]),
            int(group_dict["second"]),
        )
        return timestamp.replace(tzinfo=timezone.utc).astimezone(TIME_ZONE_NEW_YORK)

    @property
    def seconds_since_game_start(self):
        if not self.time_pitch_thrown or not self.game_start_time:
            return 0
        return int((self.time_pitch_thrown - self.game_start_time).total_seconds())

    def as_dict(self):
        """Convert pitch log to a dictionary."""
        return dict(
            __brooks_pitchfx_data__=True,
            pitcher_name=self.pitcher_name,
            pitch_app_id=self.pitch_app_id,
            pitcher_id=self.pitcher_id,
            batter_id=self.batter_id,
            pitcher_team_id_bb=self.pitcher_team_id_bb,
            opponent_team_id_bb=self.opponent_team_id_bb,
            bb_game_id=self.bb_game_id,
            bbref_game_id=self.bbref_game_id,
            table_row_number=int(self.table_row_number),
            park_sv_id=self.park_sv_id,
            play_guid=self.play_guid,
            ab_total=int(self.ab_total),
            ab_count=int(self.ab_count),
            ab_id=int(self.ab_id),
            des=self.des,
            type=self.type,
            id=int(self.id),
            sz_top=float(self.sz_top),
            sz_bot=float(self.sz_bot),
            pfx_xdatafile=float(self.pfx_xdatafile),
            pfx_zdatafile=float(self.pfx_zdatafile),
            mlbam_pitch_name=self.mlbam_pitch_name,
            zone_location=self.zone_location,
            pitch_con=float(self.pitch_con),
            stand=self.stand,
            strikes=int(self.strikes),
            balls=int(self.balls),
            p_throws=self.p_throws,
            pdes=self.pdes,
            spin=float(self.spin),
            norm_ht=float(self.norm_ht),
            inning=int(self.inning),
            tstart=float(self.tstart),
            vystart=float(self.vystart),
            ftime=float(self.ftime),
            pfx_x=float(self.pfx_x),
            pfx_z=float(self.pfx_z),
            uncorrected_pfx_x=float(self.uncorrected_pfx_x),
            uncorrected_pfx_z=float(self.uncorrected_pfx_z),
            x0=float(self.x0),
            y0=float(self.y0),
            z0=float(self.z0),
            vx0=float(self.vx0),
            vy0=float(self.vy0),
            vz0=float(self.vz0),
            ax=float(self.ax),
            ay=float(self.ay),
            az=float(self.az),
            start_speed=float(self.start_speed),
            px=float(self.px),
            pz=float(self.pz),
            pxold=float(self.pxold),
            pzold=float(self.pzold),
            tm_spin=int(self.tm_spin),
            sb=int(self.sb),
            game_start_time_str=self.game_start_time_str,
            time_pitch_thrown_str=self.get_time_pitch_thrown_str(),
            has_zone_location=self.has_zone_location,
            is_patched=self.is_patched,
            is_duplicate_guid=self.is_duplicate_guid,
            is_duplicate_pitch_number=self.is_duplicate_pitch_number,
            is_invalid_ibb=self.is_invalid_ibb,
            is_out_of_sequence=self.is_out_of_sequence,
        )

    def get_time_pitch_thrown_str(self):
        return self.time_pitch_thrown.strftime(DT_AWARE) if self.time_pitch_thrown else None



from sqlalchemy import case, func, join, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, PitchAppScrapeStatus, PitchFx


class PitchApp_PitchFx_View(Base):
    __table__ = create_view(
        name="pitch_app_pitchfx",
        selectable=select(
            [
                PitchAppScrapeStatus.id.label("id"),
                PitchAppScrapeStatus.pitcher_id_mlb.label("pitcher_id_mlb"),
                PitchAppScrapeStatus.pitch_app_id.label("pitch_app_id"),
                PitchAppScrapeStatus.bbref_game_id.label("bbref_game_id"),
                func.count(PitchFx.id).label("total_pitches"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CH", 1)], else_=0)).label("ch_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CU", 1)], else_=0)).label("cu_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "EP", 1)], else_=0)).label("ep_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FA", 1)], else_=0)).label("fa_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FC", 1)], else_=0)).label("fc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FF", 1)], else_=0)).label("ff_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FS", 1)], else_=0)).label("fs_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FT", 1)], else_=0)).label("ft_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FO", 1)], else_=0)).label("fo_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "IN", 1)], else_=0)).label("in_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KC", 1)], else_=0)).label("kc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KN", 1)], else_=0)).label("kn_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "PO", 1)], else_=0)).label("po_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SC", 1)], else_=0)).label("sc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SI", 1)], else_=0)).label("si_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SL", 1)], else_=0)).label("sl_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "UN", 1)], else_=0)).label("un_count"),
            ]
        )
        .select_from(
            join(
                PitchAppScrapeStatus,
                PitchFx,
                PitchAppScrapeStatus.id == PitchFx.pitch_app_db_id,
                isouter=True,
            )
        )
        .group_by(PitchAppScrapeStatus.id),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )



from vigorish.util.list_helpers import as_dict_list


class BBRefHalfInning:
    inning_id = ""
    inning_label = ""
    begin_inning_summary = ""
    end_inning_summary = ""
    inning_total_runs = None
    inning_total_hits = None
    inning_total_errors = None
    inning_total_left_on_base = None
    away_team_runs_after_inning = None
    home_team_runs_after_inning = None
    game_events = []
    substitutions = []
    misc_events = []

    def as_dict(self):
        """Convert half-inning to a dictionary."""
        return dict(
            __bbref_boxscore_half_inning__=True,
            inning_id=self.inning_id,
            inning_label=self.inning_label,
            begin_inning_summary=self.begin_inning_summary,
            end_inning_summary=self.end_inning_summary,
            inning_total_runs=int(self.inning_total_runs),
            inning_total_hits=int(self.inning_total_hits),
            inning_total_errors=int(self.inning_total_errors),
            inning_total_left_on_base=int(self.inning_total_left_on_base),
            away_team_runs_after_inning=int(self.away_team_runs_after_inning),
            home_team_runs_after_inning=int(self.home_team_runs_after_inning),
            game_events=as_dict_list(self.game_events),
            substitutions=as_dict_list(self.substitutions),
            misc_events=as_dict_list(self.misc_events),
        )



from datetime import datetime

from vigorish.util.dt_format_strings import DATE_ONLY, DT_AWARE


def dict_from_dataclass(obj, data_class, date_format=DT_AWARE):
    obj_dict = {}
    for name, field in data_class.__dataclass_fields__.items():
        value = getattr(obj, name, None)
        if field.type is int:
            obj_dict[name] = int(value)
        elif field.type is bool:
            obj_dict[name] = True if value else False
        elif not value:
            obj_dict[name] = None
        elif field.type is datetime:
            obj_dict[name] = value.strftime(date_format)
        elif field.type is float:
            obj_dict[name] = float(value)
        else:
            obj_dict[name] = value
    return obj_dict


def sanitize_row_dict(row_dict, date_format=DATE_ONLY):
    return [val_to_string(val, date_format) for val in row_dict.values()]


def val_to_string(val, date_format):
    return (
        val
        if isinstance(val, str)
        else val.strftime(date_format).strip()
        if isinstance(val, datetime)
        else 0
        if (isinstance(val, bool) and not val)
        else 1
        if (isinstance(val, bool) and val)
        else ""
        if not val
        else str(val)
    )



"""Constant values that are referenced from multiple places."""
from bullet import colors

from vigorish.enums import DataSet, VigFile

ENV_VAR_NAMES = [
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_DEFAULT_REGION",
    "CONFIG_FILE",
    "DATABASE_URL",
]

BR_BB_TEAM_ID_MAP = {
    "CHW": "CHA",
    "CHC": "CHN",
    "KCR": "KCA",
    "LAA": "ANA",
    "LAD": "LAN",
    "NYY": "NYA",
    "NYM": "NYN",
    "SDP": "SDN",
    "SFG": "SFN",
    "STL": "SLN",
    "TBR": "TBA",
    "WSN": "WAS",
}

BB_BR_TEAM_ID_MAP = {v: k for k, v in BR_BB_TEAM_ID_MAP.items()}

DEFENSE_POSITIONS = [
    "P",
    "C",
    "1B",
    "2B",
    "3B",
    "SS",
    "RF",
    "CF",
    "LF",
    "DH",
    "PH",
    "PR",
]

PITCH_TYPE_DICT = dict(
    CH="Changeup",
    CU="Curveball",
    EP="Eephus",
    FA="Fastball",
    FC="Cutter",
    FF="Four-seam Fastball",
    FS="Splitter",
    FT="Two-seam Fastball",
    FO="Forkball",
    IN="Intent ball",
    KC="Knuckle ball Curve",
    KN="Knuckle ball",
    PO="Pitch Out",
    SC="Screwball",
    SI="Sinker",
    SL="Slider",
    UN="Unknown",
)

AT_BAT_RESULTS_ALL = [
    "Bunt Groundout",
    "Bunt Lineout",
    "Bunt Pop Out",
    "Double",
    "Double Play",
    "Field Error",
    "Fielders Choice",
    "Fielders Choice Out",
    "Flyout",
    "Forceout",
    "Grounded Into DP",
    "Groundout",
    "Hit By Pitch",
    "Home Run",
    "Intent Walk",
    "Lineout",
    "Pop Out",
    "Runner Out",
    "Sac Bunt",
    "Sac Fly",
    "Sac Fly DP",
    "Single",
    "Strikeout",
    "Strikeout - DP",
    "Triple",
    "Walk",
]

PPB_PITCH_LOG_DICT = {
    "C": {"description": "Called strike", "pitch_counts": 1},
    "S": {"description": "Swinging strike", "pitch_counts": 1},
    "F": {"description": "Foul", "pitch_counts": 1},
    "B": {"description": "Ball", "pitch_counts": 1},
    "X": {"description": "Ball put into play by batter", "pitch_counts": 1},
    "T": {"description": "Foul tip", "pitch_counts": 1},
    "K": {"description": "Strike (unknown type)", "pitch_counts": 1},
    "I": {"description": "Intentional ball", "pitch_counts": 1},
    "H": {"description": "Hit batter", "pitch_counts": 1},
    "L": {"description": "Foul bunt", "pitch_counts": 1},
    "M": {"description": "Missed bunt attempt", "pitch_counts": 1},
    "N": {"description": "No pitch (on balks and interference calls)", "pitch_counts": 0},
    "O": {"description": "Foul tip on bunt", "pitch_counts": 1},
    "P": {"description": "Pitchout", "pitch_counts": 1},
    "Q": {"description": "Swinging on pitchout", "pitch_counts": 1},
    "R": {"description": "Foul ball on pitchout", "pitch_counts": 1},
    "U": {"description": "Unknown or missed pitch", "pitch_counts": 1},
    "V": {"description": "(Called ball because pitcher went to his mouth)", "pitch_counts": 0},
    "Y": {"description": "Ball put into play on pitchout", "pitch_counts": 1},
    "1": {"description": "(Pickoff throw to first)", "pitch_counts": 0},
    "2": {"description": "(Pickoff throw to second)", "pitch_counts": 0},
    "3": {"description": "(Pickoff throw to third)", "pitch_counts": 0},
    ">": {"description": "(Runner going on the pitch)", "pitch_counts": 0},
    "+": {"description": "(Pickoff throw by the catcher)", "pitch_counts": 0},
    "*": {"description": "(The following pitch was blocked by the catcher)", "pitch_counts": 0},
    ".": {"description": "(Play not involving the batter)", "pitch_counts": 0},
}

VENUE_TERMS = ["stadium", "park", "field", "coliseum", "centre", "estadio", "dome"]

JOB_SPINNER_COLORS = {
    DataSet.BBREF_GAMES_FOR_DATE: "red",
    DataSet.BROOKS_GAMES_FOR_DATE: "blue",
    DataSet.BBREF_BOXSCORES: "green",
    DataSet.BROOKS_PITCH_LOGS: "cyan",
    DataSet.BROOKS_PITCHFX: "magenta",
}

DATA_SET_NAMES_LONG = {
    "Games for Date (bbref.com)": DataSet.BBREF_GAMES_FOR_DATE,
    "Games for Date (brooksbaseball.net)": DataSet.BROOKS_GAMES_FOR_DATE,
    "Boxscores (bbref.com)": DataSet.BBREF_BOXSCORES,
    "Pitch Logs for Game (brooksbaseball.net)": DataSet.BROOKS_PITCH_LOGS,
    "PitchFX Logs (brooksbaseball.net)": DataSet.BROOKS_PITCHFX,
}

FILE_TYPE_NAME_MAP = {str(ft): ft for ft in VigFile}
DATA_SET_NAME_MAP = {str(ds): ds for ds in DataSet}

CLI_COLORS = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
]

BULLET_COLORS = {
    "black": colors.foreground["black"],
    "red": colors.foreground["red"],
    "green": colors.foreground["green"],
    "yellow": colors.foreground["yellow"],
    "blue": colors.foreground["blue"],
    "magenta": colors.foreground["magenta"],
    "cyan": colors.foreground["cyan"],
    "white": colors.foreground["white"],
    "bright_black": colors.bright(colors.foreground["black"]),
    "bright_red": colors.bright(colors.foreground["red"]),
    "bright_green": colors.bright(colors.foreground["green"]),
    "bright_yellow": colors.bright(colors.foreground["yellow"]),
    "bright_blue": colors.bright(colors.foreground["blue"]),
    "bright_magenta": colors.bright(colors.foreground["magenta"]),
    "bright_cyan": colors.bright(colors.foreground["cyan"]),
    "bright_white": colors.bright(colors.foreground["white"]),
}

EMOJI_DICT = dict(
    HAND_POINTER="👉",
    BOOKMARK="🔖",
    FOLDER="📁",
    CHART="📊",
    BOOK="📚",
    TABBED_FILES="🗂️",
    TOOLS="🛠️ ",
    KNIFE="🔪",
    ROBOT="🤖",
    KNOBS="🎛️",
    BASKET="🧺",
    PACKAGE="📦",
    BAGS="🛍️",
    HONEY_POT="🍯",
    CLOCK="🕚",
    PAGER="📟",
    SHRUG="🤷‍♂️",
    CRYING="😢",
    WEARY="😩",
    UPSIDE_DOWN="🙃",
    CONFUSED="😕",
    THUMBS_UP="👍",
    CLOUD="🌧",
    SPIRAL="🌀",
    DIZZY="💫",
    BOLT="⚡",
    BOMB="💣",
    FLASHLIGHT="🔦",
    MICROSCOPE="🔬",
    NEWSPAPER="📰",
    CAP="🧢",
    BASEBALL="⚾",
    SOFTBALL="🥎",
    HOTDOG="🌭",
    CIRCLE="🔵",
    GEAR="⚙️ ",
    BANG="💥",
    BLUE_DIAMOND="🔹",
    CHECK="☑️",
    ASTERISK="✳️",
    UP="⬆️",
    DOWN="⬇️",
    LEFT="⬅️",
    RIGHT="➡️",
    PASSED="✅",
    FAILED="❌",
    QUESTION="❔",
    COOL="🆒",
    BACK="👈",
    EXIT="❎",
)

MENU_NUMBERS = {
    1: "1️⃣",
    2: "2️⃣",
    3: "3️⃣",
    4: "4️⃣",
    5: "5️⃣",
    6: "6️⃣",
    7: "7️⃣",
    8: "8️⃣",
    9: "9️⃣",
    10: "🔟",
}

FAKE_SPINNER = {"interval": 140, "frames": ["⚡", "⚡"]}

FIGLET_FONTS = [
    "3-d",
    "acrobatic",
    "banner3-D",
    "basic",
    "charact6",
    "contrast",
    "cyberlarge",
    "doom",
    "double",
    "ebbs_1__",
    "eftiwater",
    "epic",
    "fender",
    "graffiti",
    "home_pak",
    "isometric1",
    "jazmine",
    "kban",
    "larry3d",
    "lean",
    "lockergnome",
    "nancyj-fancy",
    "os2",
    "peaks",
    "rowancap",
    "sblood",
    "shimrod",
    "slant",
    "smscript",
    "space_op",
    "speed",
    "stellar",
    "ticksslant",
    "tinker-toy",
    "tubular",
    "usaflag",
]



import os
from pathlib import Path

from dotenv import load_dotenv

from vigorish.config.database import SQLITE_PROD_URL
from vigorish.constants import ENV_VAR_NAMES
from vigorish.util.result import Result

VIG_FOLDER = Path.home() / ".vig"
DOTENV_FILE = VIG_FOLDER / ".env"
DEFAULT_CONFIG = VIG_FOLDER / "vig.config.json"


def create_default_dotenv_file(dotenv_file, config_file=None, db_url=None):
    env_var_dict = {var_name: "" for var_name in ENV_VAR_NAMES}
    env_var_dict["CONFIG_FILE"] = config_file if config_file else DEFAULT_CONFIG
    env_var_dict["DATABASE_URL"] = db_url if db_url else SQLITE_PROD_URL
    env_var_strings = [f"{name}={value}" for name, value in env_var_dict.items()]
    dotenv_file.write_text("\n".join(env_var_strings))


class DotEnvFile:
    def __init__(self, dotenv_filepath=None):
        if os.environ.get("ENV") == "TEST":
            self.dotenv_filepath = Path(os.environ.get("DOTENV_FILE"))
        else:
            self.dotenv_filepath = dotenv_filepath if dotenv_filepath else DOTENV_FILE
        self.env_var_dict = self.read_dotenv_file()

    def read_dotenv_file(self):
        """Parse .env file to a dictionary of environment variables."""
        if not self.dotenv_filepath.exists():
            create_default_dotenv_file(dotenv_file=self.dotenv_filepath)
        if not self.dotenv_filepath.is_file():
            raise TypeError(f"Unable to open file: {self.dotenv_filepath}")
        load_dotenv(self.dotenv_filepath)
        file_text = self.dotenv_filepath.read_text()
        env_var_split = [f for f in file_text.split("\n") if f and "=" in f]
        env_var_split = [s.split("=") for s in env_var_split]
        return {v[0]: v[1].strip('"').strip("'").strip() for v in env_var_split if len(v) == 2}

    def get_current_value(self, env_var_name):
        if env_var_name not in ENV_VAR_NAMES:
            return None
        return self.env_var_dict.get(env_var_name, None)

    def change_value(self, env_var_name, new_value):
        if env_var_name not in ENV_VAR_NAMES:
            return Result.Fail(f"{env_var_name} is not a recognized environment variable.")
        self.env_var_dict[env_var_name] = new_value
        self.write_dotenv_file()
        self.env_var_dict = self.read_dotenv_file()
        return Result.Ok()

    def restart_required_on_change(self, setting_name):
        return setting_name in ["CONFIG_FILE", "DATABASE_URL"]

    def write_dotenv_file(self):
        env_var_strings = [f"{name}={value}" for name, value in self.env_var_dict.items()]
        self.dotenv_filepath.write_text("\n".join(env_var_strings))



from vigorish.config.database import DateScrapeStatus
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.bbref_games_for_date.parse_html import parse_bbref_dashboard_page
from vigorish.scrape.scrape_task import ScrapeTaskABC
from vigorish.status.update_status_bbref_games_for_date import (
    update_bbref_games_for_date_single_date,
)
from vigorish.util.result import Result


class ScrapeBBRefGamesForDate(ScrapeTaskABC):
    def __init__(self, db_job, db_session, config, scraped_data):
        self.data_set = DataSet.BBREF_GAMES_FOR_DATE
        super().__init__(db_job, db_session, config, scraped_data)

    def check_prerequisites(self, game_date):
        return Result.Ok()

    def check_current_status(self, game_date):
        bbref_games_for_date = DateScrapeStatus.verify_bbref_daily_dashboard_scraped_for_date(
            self.db_session, game_date
        )
        if bbref_games_for_date and self.scrape_condition == ScrapeCondition.ONLY_MISSING_DATA:
            return Result.Fail("skip")
        return Result.Ok()

    def parse_html(self, html, url_id, url):
        return parse_bbref_dashboard_page(html, url_id, url)

    def update_status(self, game_date, parsed_data):
        return update_bbref_games_for_date_single_date(self.db_session, self.season, parsed_data)



from dataclasses import dataclass

from dataclass_csv import accept_whitespaces
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from vigorish.config.database import Base
from vigorish.util.dataclass_helpers import dict_from_dataclass, sanitize_row_dict
from vigorish.util.dt_format_strings import DATE_ONLY_TABLE_ID
from vigorish.util.list_helpers import display_dict
from vigorish.util.pitch_calcs import calc_pitch_mix


class PitchAppScrapeStatus(Base):

    __tablename__ = "scrape_status_pitch_app"
    id = Column(Integer, primary_key=True)
    pitcher_id_mlb = Column(Integer)
    pitch_app_id = Column(String, unique=True)
    bbref_game_id = Column(String)
    bb_game_id = Column(String)
    scraped_pitchfx = Column(Integer, default=0)
    no_pitchfx_data = Column(Integer, default=0)
    combined_pitchfx_bbref_data = Column(Integer, default=0)
    imported_pitchfx = Column(Integer, default=0)
    invalid_pitchfx = Column(Integer, default=0)
    pitchfx_error = Column(Integer, default=0)
    pitch_count_pitch_log = Column(Integer, default=0)
    pitch_count_bbref = Column(Integer, default=0)
    pitch_count_pitchfx = Column(Integer, default=0)
    pitch_count_pitchfx_audited = Column(Integer, default=0)
    patched_pitchfx_count = Column(Integer, default=0)
    missing_pitchfx_count = Column(Integer, default=0)
    extra_pitchfx_count = Column(Integer, default=0)
    invalid_pitchfx_count = Column(Integer, default=0)
    extra_pitchfx_removed_count = Column(Integer, default=0)
    duplicate_guid_removed_count = Column(Integer, default=0)
    batters_faced_bbref = Column(Integer, default=0)
    batters_faced_pitchfx = Column(Integer, default=0)
    total_at_bats_pitchfx_complete = Column(Integer, default=0)
    total_at_bats_patched_pitchfx = Column(Integer, default=0)
    total_at_bats_missing_pitchfx = Column(Integer, default=0)
    total_at_bats_extra_pitchfx = Column(Integer, default=0)
    total_at_bats_extra_pitchfx_removed = Column(Integer, default=0)
    total_at_bats_pitchfx_error = Column(Integer, default=0)
    total_at_bats_invalid_pitchfx = Column(Integer, default=0)

    pitcher_id = Column(Integer, ForeignKey("player.id"), index=True)
    scrape_status_game_id = Column(Integer, ForeignKey("scrape_status_game.id"), index=True)
    scrape_status_date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))
    pitch_mix_view = relationship(
        "PitchApp_PitchFx_View",
        backref="original",
        uselist=False,
        primaryjoin="PitchAppScrapeStatus.id==PitchApp_PitchFx_View.id",
        foreign_keys="PitchApp_PitchFx_View.id",
    )

    @hybrid_property
    def pitch_mix(self):
        return calc_pitch_mix(self.pitch_mix_view.__dict__, self.pitch_mix_view.total_pitches)

    @hybrid_property
    def contains_patched_data(self):
        return self.patched_pitchfx_count > 0

    @hybrid_property
    def contains_only_patched_data(self):
        return self.patched_pitchfx_count > 0 and (
            self.pitch_count_pitchfx_audited == self.patched_pitchfx_count
        )

    def __repr__(self):
        return f'<PitchAppScrapeStatus pitch_app_id="{self.pitch_app_id}">'

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def display(self):
        title = f"SCRAPE STATUS FOR PITCH APPEARANCE: {self.pitch_app_id}"
        display_dict(self.as_dict(), title=title)

    def as_csv_dict(self):
        return dict_from_dataclass(self, PitchAppScrapeStatusCsvRow)

    @classmethod
    def get_csv_col_names(cls):
        return [name for name in PitchAppScrapeStatusCsvRow.__dataclass_fields__.keys()]

    @classmethod
    def export_table_as_csv(cls, db_session):
        col_names = ",".join(cls.get_csv_col_names())
        csv_dicts = (obj.as_csv_dict() for obj in db_session.query(cls).all())
        csv_rows = (",".join(sanitize_row_dict(d)) for d in csv_dicts)
        yield col_names
        for row in csv_rows:
            yield row

    @classmethod
    def find_by_pitch_app_id(cls, db_session, pitch_app_id):
        return db_session.query(cls).filter_by(pitch_app_id=pitch_app_id).first()

    @classmethod
    def get_all_scraped_pitch_app_ids_for_season(cls, db_session, season_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if pitch_app_status.scraped_pitchfx == 1
        ]

    @classmethod
    def get_all_unscraped_pitch_app_ids_for_season(cls, db_session, season_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if pitch_app_status.scraped_pitchfx == 0
        ]

    @classmethod
    def get_pitch_app_ids_without_pitchfx_data(cls, db_session, season_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if pitch_app_status.no_pitchfx_data == 1
        ]

    @classmethod
    def get_all_pitch_app_ids(cls, db_session, season_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(season_id=season_id).all()
        ]

    @classmethod
    def get_all_pitch_app_ids_for_date(cls, db_session, date):
        date_id = date.strftime(DATE_ONLY_TABLE_ID)
        return [
            pitch_app.pitch_app_id
            for pitch_app in db_session.query(cls)
            .filter_by(scrape_status_date_id=int(date_id))
            .all()
        ]

    @classmethod
    def get_all_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .all()
        ]

    @classmethod
    def get_all_scraped_pitch_app_ids_for_game_with_pitchfx_data(cls, db_session, bbref_game_id):
        pitch_apps_with_pfx_data = [
            pitch_app_status
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=1)
            .filter_by(no_pitchfx_data=0)
            .all()
        ]
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in pitch_apps_with_pfx_data
            if not pitch_app_status.contains_only_patched_data
        ]

    @classmethod
    def get_all_unscraped_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=0)
            .all()
        ]

    @classmethod
    def get_all_scraped_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=1)
            .all()
        ]

    @classmethod
    def get_pitch_app_id_map(cls, db_session):
        all_pitch_apps = db_session.query(cls).all()
        return {pa.pitch_app_id: pa.id for pa in all_pitch_apps}


@accept_whitespaces
@dataclass
class PitchAppScrapeStatusCsvRow:
    id: int
    pitcher_id_mlb: int = 0
    pitch_app_id: str = None
    bbref_game_id: str = None
    bb_game_id: str = None
    scraped_pitchfx: int = 0
    no_pitchfx_data: int = 0
    combined_pitchfx_bbref_data: int = 0
    imported_pitchfx: int = 0
    invalid_pitchfx: int = 0
    pitchfx_error: int = 0
    pitch_count_pitch_log: int = 0
    pitch_count_bbref: int = 0
    pitch_count_pitchfx: int = 0
    pitch_count_pitchfx_audited: int = 0
    patched_pitchfx_count: int = 0
    missing_pitchfx_count: int = 0
    extra_pitchfx_count: int = 0
    invalid_pitchfx_count: int = 0
    extra_pitchfx_removed_count: int = 0
    duplicate_guid_removed_count: int = 0
    batters_faced_bbref: int = 0
    batters_faced_pitchfx: int = 0
    total_at_bats_pitchfx_complete: int = 0
    total_at_bats_patched_pitchfx: int = 0
    total_at_bats_missing_pitchfx: int = 0
    total_at_bats_extra_pitchfx: int = 0
    total_at_bats_extra_pitchfx_removed: int = 0
    total_at_bats_pitchfx_error: int = 0
    total_at_bats_invalid_pitchfx: int = 0



"""Scrape brooksbaseball daily dashboard page."""
import uuid
from collections import defaultdict, OrderedDict
from string import Template

from lxml import html

from vigorish.scrape.brooks_pitchfx.models.pitchfx import BrooksPitchFxData
from vigorish.scrape.brooks_pitchfx.models.pitchfx_log import BrooksPitchFxLog
from vigorish.util.dt_format_strings import DT_AWARE
from vigorish.util.result import Result

PITCHFX_COLUMN_NAMES = "//tr/th/text()"
PITCHFX_TABLE_ROWS = '//tr[@style="white-space:nowrap"]'
T_PITCHFX_MEASUREMENT = "td[${i}]//text()"
FILTER_NAMES = ["datestamp", "gid", "pitcher_team"]


def parse_pitchfx_log(scraped_html, pitch_log):
    page_content = html.fromstring(scraped_html, base_url=pitch_log.pitchfx_url)
    pitchfx_log_dict = {}
    result = parse_pitchfx_table(page_content, pitch_log)
    if result.failure:
        return result
    pitchfx_data = result.value
    pitchfx_log_dict["pitchfx_log"] = pitchfx_data
    pitchfx_log_dict["pitch_count_by_inning"] = get_pitch_count_by_inning(pitchfx_data)
    pitchfx_log_dict["total_pitch_count"] = len(pitchfx_data)
    pitchfx_log_dict["pitcher_name"] = pitch_log.pitcher_name
    pitchfx_log_dict["pitcher_id_mlb"] = pitch_log.pitcher_id_mlb
    pitchfx_log_dict["pitch_app_id"] = pitch_log.pitch_app_id
    pitchfx_log_dict["pitcher_team_id_bb"] = pitch_log.pitcher_team_id_bb
    pitchfx_log_dict["opponent_team_id_bb"] = pitch_log.opponent_team_id_bb
    pitchfx_log_dict["bb_game_id"] = pitch_log.bb_game_id
    pitchfx_log_dict["bbref_game_id"] = pitch_log.bbref_game_id
    pitchfx_log_dict["game_date_year"] = pitch_log.game_date_year
    pitchfx_log_dict["game_date_month"] = pitch_log.game_date_month
    pitchfx_log_dict["game_date_day"] = pitch_log.game_date_day
    pitchfx_log_dict["game_time_hour"] = pitch_log.game_time_hour
    pitchfx_log_dict["game_time_minute"] = pitch_log.game_time_minute
    pitchfx_log_dict["time_zone_name"] = pitch_log.time_zone_name
    pitchfx_log_dict["pitchfx_url"] = pitch_log.pitchfx_url
    pitchfx_log = BrooksPitchFxLog(**pitchfx_log_dict)
    return Result.Ok(pitchfx_log)


def parse_pitchfx_table(page_content, pitch_log):
    pitchfx_data = []
    column_names = page_content.xpath(PITCHFX_COLUMN_NAMES)
    table_rows = page_content.xpath(PITCHFX_TABLE_ROWS)
    for i, row in enumerate(table_rows):
        result = parse_pitchfx_data(column_names, row, i, pitch_log)
        if result.failure:
            return result
        pitchfx = result.value
        pitchfx_data.append(pitchfx)
    return fix_missing_des(pitchfx_data, pitch_log)


def parse_pitchfx_data(column_names, table_row, row_num, pitch_log):
    pitchfx_dict = {}
    for i, name in enumerate(column_names):
        if name.lower() in FILTER_NAMES:
            continue
        query = Template(T_PITCHFX_MEASUREMENT).substitute(i=(i + 1))
        results = table_row.xpath(query)
        if not results:
            if name == "zone_location":
                pitchfx_dict["zone_location"] = 99
                continue
            if name == "des":
                pitchfx_dict["des"] = "missing_des"
                continue
            if name == "pdes":
                pitchfx_dict["pdes"] = "missing_pdes"
                continue
            if name == "play_guid":
                pitchfx_dict["play_guid"] = str(uuid.uuid4())
                continue
            error = (
                f"Error occurred attempting to parse '{name}' (column #{i}) from pitchfx table:\n"
                f"Game ID.......: {pitch_log.bbref_game_id}\n"
                f"Pitcher.......: {pitch_log.pitcher_name} ({pitch_log.pitcher_id_mlb})\n"
                f"Row Number....: {row_num}\n"
                f"PitchFX URL...: {pitch_log.pitchfx_url}\n"
                f"Partial Data..: {pitchfx_dict}"
            )
            return Result.Fail(error)
        pitchfx_dict[name.lower()] = results[0]
    pitchfx_dict["pitcher_name"] = pitch_log.pitcher_name
    pitchfx_dict["pitch_app_id"] = pitch_log.pitch_app_id
    pitchfx_dict["pitcher_team_id_bb"] = pitch_log.pitcher_team_id_bb
    pitchfx_dict["opponent_team_id_bb"] = pitch_log.opponent_team_id_bb
    pitchfx_dict["bb_game_id"] = pitch_log.bb_game_id
    pitchfx_dict["bbref_game_id"] = pitch_log.bbref_game_id
    pitchfx_dict["table_row_number"] = row_num
    pitchfx = BrooksPitchFxData(**pitchfx_dict)

    game_start_time = pitch_log.game_start_time
    pitchfx.game_start_time_str = game_start_time.strftime(DT_AWARE) if game_start_time else ""
    pitchfx.has_zone_location = True if pitchfx_dict["zone_location"] != 99 else False
    return Result.Ok(pitchfx)


def fix_missing_des(pitchfx_data, pitch_log):
    missing_des = any(pfx.des == "missing_des" for pfx in pitchfx_data)
    if not missing_des:
        return Result.Ok(pitchfx_data)
    fix_ab_ids = list(set([pfx.ab_id for pfx in pitchfx_data if pfx.des == "missing_des"]))
    for ab_id in fix_ab_ids:
        missing_des_this_ab = [
            pfx for pfx in pitchfx_data if pfx.ab_id == ab_id and pfx.des == "missing_des"
        ]
        valid_des_this_ab = list(
            set(
                [
                    pfx.des
                    for pfx in pitchfx_data
                    if pfx.ab_id == ab_id and not pfx.des == "missing_des"
                ]
            )
        )
        if len(valid_des_this_ab) == 0:
            continue
        if len(valid_des_this_ab) == 1:
            for pfx in missing_des_this_ab:
                pfx.des = valid_des_this_ab[0]
            continue
        error = (
            f"Unable to fix missing description for pitchfx ab_id={ab_id}\n"
            f"Game ID.......: {pitch_log.bbref_game_id}\n"
            f"Pitcher.......: {pitch_log.pitcher_name} ({pitch_log.pitcher_id_mlb})\n"
            f"PitchFX URL...: {pitch_log.pitchfx_url}\n"
            f"des values....: {valid_des_this_ab}"
        )
        return Result.Fail(error)
    return Result.Ok(pitchfx_data)


def get_pitch_count_by_inning(pitchfx_data):
    pitch_count_unordered = defaultdict(int)
    for pfx in pitchfx_data:
        pitch_count_unordered[pfx.inning] += 1
    pitch_count_ordered = OrderedDict()
    for k in sorted(pitch_count_unordered.keys()):
        pitch_count_ordered[k] = pitch_count_unordered[k]
    return pitch_count_ordered



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, GameScrapeStatus, Season


class Season_Game_View(Base):
    __table__ = create_view(
        name="season_game_status",
        selectable=select(
            [
                Season.id.label("id"),
                func.count(GameScrapeStatus.id).label("total_games"),
                func.sum(GameScrapeStatus.scraped_bbref_boxscore).label(
                    "total_scraped_bbref_boxscore"
                ),
                func.sum(GameScrapeStatus.scraped_brooks_pitch_logs).label(
                    "total_scraped_brooks_pitch_logs"
                ),
                func.sum(GameScrapeStatus.combined_data_success).label(
                    "total_combined_data_success"
                ),
                func.sum(GameScrapeStatus.combined_data_fail).label("total_combined_data_fail"),
                func.sum(GameScrapeStatus.pitch_app_count_bbref).label(
                    "total_pitch_app_count_bbref"
                ),
                func.sum(GameScrapeStatus.pitch_app_count_brooks).label(
                    "total_pitch_app_count_brooks"
                ),
                func.sum(GameScrapeStatus.total_pitch_count_bbref).label("total_pitch_count_bbref"),
            ]
        )
        .select_from(
            join(
                Season,
                GameScrapeStatus,
                Season.id == GameScrapeStatus.season_id,
                isouter=True,
            )
        )
        .group_by(Season.id),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )



"""Database file and connection details."""
# flake8: noqa
import os
from pathlib import Path

from sqlalchemy import create_engine, func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

Base = declarative_base()

from vigorish.config.project_paths import CSV_FOLDER, VIG_FOLDER
from vigorish.models.pitchfx import PitchFx
from vigorish.models.player import Player
from vigorish.models.player_id import PlayerId
from vigorish.models.scrape_error import ScrapeError
from vigorish.models.scrape_job import ScrapeJob
from vigorish.models.season import Season
from vigorish.models.status_date import DateScrapeStatus
from vigorish.models.status_game import GameScrapeStatus
from vigorish.models.status_pitch_appearance import PitchAppScrapeStatus
from vigorish.models.team import Team
from vigorish.models.time_between_pitches import TimeBetweenPitches
from vigorish.models.views.date_pitch_app_view import Date_PitchApp_View
from vigorish.models.views.game_pitch_app_view import Game_PitchApp_View
from vigorish.models.views.pitch_app_pitchfx_view import PitchApp_PitchFx_View
from vigorish.models.views.player_pitchfx_view import (
    Pitch_Mix_All_View,
    Pitch_Mix_By_Year_View,
    Pitch_Mix_Left_View,
    Pitch_Mix_Right_View,
)
from vigorish.models.views.season_date_view import Season_Date_View
from vigorish.models.views.season_game_pitch_app_view import Season_Game_PitchApp_View
from vigorish.models.views.season_game_view import Season_Game_View
from vigorish.models.views.season_pitch_app_view import Season_PitchApp_View
from vigorish.setup.populate_tables import populate_tables, populate_tables_for_restore
from vigorish.util.result import Result

SQLITE_DEV_URL = f"sqlite:///{VIG_FOLDER.joinpath('vig_dev.db')}"
SQLITE_PROD_URL = f"sqlite:///{VIG_FOLDER.joinpath('vig.db')}"


def get_db_url():
    if os.environ.get("ENV") == "TEST":
        return os.environ.get("DATABASE_URL")
    db_url = os.getenv("DATABASE_URL", "")
    if db_url and db_url.startswith("/"):
        db_url = f"sqlite:///{db_url}"
    env = os.getenv("ENV", "prod")
    return db_url if db_url else SQLITE_DEV_URL if env == "dev" else SQLITE_PROD_URL


def initialize_database(app, csv_folder=None):
    if not csv_folder:
        csv_folder = CSV_FOLDER
    Base.metadata.drop_all(app["db_engine"])
    Base.metadata.create_all(app["db_engine"])
    return populate_tables(app, csv_folder)


def prepare_database_for_restore(app, csv_folder=None):
    if not csv_folder:
        csv_folder = CSV_FOLDER
    Base.metadata.drop_all(app["db_engine"])
    Base.metadata.create_all(app["db_engine"])
    return populate_tables_for_restore(app, csv_folder)


def reset_database_connection(app, db_url=None):
    if not db_url:
        db_url = get_db_url()
    result = delete_sqlite_database(db_url)
    if result.failure:
        return result
    app = update_app_instance(app, db_url)
    return Result.Ok(app)


def delete_sqlite_database(db_url):
    db_file = Path(db_url.replace("sqlite:///", ""))
    if not db_file.exists():
        return Result.Fail("Error occurred attempting to delete existing database file!")
    db_file.unlink()
    return Result.Ok()


def update_app_instance(app, db_url):
    app["db_session"].close()
    app["db_session"] = None
    db_engine = create_engine(db_url)
    session_maker = sessionmaker(bind=db_engine)
    db_session = session_maker()
    dotenv = app["dotenv"]
    config = app["config"]
    scraped_data = app["scraped_data"]
    scraped_data.db_engine = db_engine
    scraped_data.db_session = db_session
    return {
        "dotenv": dotenv,
        "config": config,
        "db_engine": db_engine,
        "db_session": db_session,
        "scraped_data": scraped_data,
    }


def db_setup_complete(db_engine, db_session):
    tables_missing = (
        "player" not in db_engine.table_names()
        or "season" not in db_engine.table_names()
        or "team" not in db_engine.table_names()
    )
    if tables_missing:
        return False
    return (
        len(db_session.query(Season).all()) > 0
        and len(db_session.query(Player).all()) > 0
        and len(db_session.query(Team).all()) > 0
    )


def get_total_number_of_rows(db_session, db_table):
    q = db_session.query(db_table)
    count_q = q.statement.with_only_columns([func.count()]).order_by(None)
    count = q.session.execute(count_q).scalar()
    return count



from sqlalchemy import case, func, join, or_, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, PitchFx, PlayerId


class Pitch_Mix_All_View(Base):
    __table__ = create_view(
        name="pitch_mix_all",
        selectable=select(
            [
                PlayerId.db_player_id.label("id"),
                PlayerId.mlb_id.label("mlb_id"),
                PlayerId.mlb_name.label("name"),
                func.count(PitchFx.id).label("total_pitches"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CH", 1)], else_=0)).label("ch_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CU", 1)], else_=0)).label("cu_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "EP", 1)], else_=0)).label("ep_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FA", 1)], else_=0)).label("fa_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FC", 1)], else_=0)).label("fc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FF", 1)], else_=0)).label("ff_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FS", 1)], else_=0)).label("fs_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FT", 1)], else_=0)).label("ft_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FO", 1)], else_=0)).label("fo_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "IN", 1)], else_=0)).label("in_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KC", 1)], else_=0)).label("kc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KN", 1)], else_=0)).label("kn_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "PO", 1)], else_=0)).label("po_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SC", 1)], else_=0)).label("sc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SI", 1)], else_=0)).label("si_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SL", 1)], else_=0)).label("sl_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "UN", 1)], else_=0)).label("un_count"),
            ]
        )
        .where(or_(PitchFx.stand == "L", PitchFx.stand == "R"))
        .select_from(
            join(
                PlayerId,
                PitchFx,
                PlayerId.db_player_id == PitchFx.pitcher_id,
                isouter=True,
            )
        )
        .group_by(PlayerId.db_player_id)
        .order_by(PlayerId.mlb_name),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )


class Pitch_Mix_Left_View(Base):
    __table__ = create_view(
        name="pitch_mix_left",
        selectable=select(
            [
                PlayerId.db_player_id.label("id"),
                PlayerId.mlb_id.label("mlb_id"),
                PlayerId.mlb_name.label("name"),
                func.count(PitchFx.id).label("total_pitches"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CH", 1)], else_=0)).label("ch_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CU", 1)], else_=0)).label("cu_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "EP", 1)], else_=0)).label("ep_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FA", 1)], else_=0)).label("fa_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FC", 1)], else_=0)).label("fc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FF", 1)], else_=0)).label("ff_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FS", 1)], else_=0)).label("fs_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FT", 1)], else_=0)).label("ft_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FO", 1)], else_=0)).label("fo_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "IN", 1)], else_=0)).label("in_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KC", 1)], else_=0)).label("kc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KN", 1)], else_=0)).label("kn_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "PO", 1)], else_=0)).label("po_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SC", 1)], else_=0)).label("sc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SI", 1)], else_=0)).label("si_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SL", 1)], else_=0)).label("sl_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "UN", 1)], else_=0)).label("un_count"),
            ]
        )
        .where(PitchFx.stand == "L")
        .select_from(
            join(
                PlayerId,
                PitchFx,
                PlayerId.db_player_id == PitchFx.pitcher_id,
                isouter=True,
            )
        )
        .group_by(PlayerId.db_player_id)
        .order_by(PlayerId.mlb_name),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )


class Pitch_Mix_Right_View(Base):
    __table__ = create_view(
        name="pitch_mix_right",
        selectable=select(
            [
                PlayerId.db_player_id.label("id"),
                PlayerId.mlb_id.label("mlb_id"),
                PlayerId.mlb_name.label("name"),
                func.count(PitchFx.id).label("total_pitches"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CH", 1)], else_=0)).label("ch_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CU", 1)], else_=0)).label("cu_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "EP", 1)], else_=0)).label("ep_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FA", 1)], else_=0)).label("fa_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FC", 1)], else_=0)).label("fc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FF", 1)], else_=0)).label("ff_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FS", 1)], else_=0)).label("fs_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FT", 1)], else_=0)).label("ft_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FO", 1)], else_=0)).label("fo_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "IN", 1)], else_=0)).label("in_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KC", 1)], else_=0)).label("kc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KN", 1)], else_=0)).label("kn_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "PO", 1)], else_=0)).label("po_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SC", 1)], else_=0)).label("sc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SI", 1)], else_=0)).label("si_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SL", 1)], else_=0)).label("sl_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "UN", 1)], else_=0)).label("un_count"),
            ]
        )
        .where(PitchFx.stand == "R")
        .select_from(
            join(
                PlayerId,
                PitchFx,
                PlayerId.db_player_id == PitchFx.pitcher_id,
                isouter=True,
            )
        )
        .group_by(PlayerId.db_player_id)
        .order_by(PlayerId.mlb_name),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )


class Pitch_Mix_By_Year_View(Base):
    __table__ = create_view(
        name="pitch_mix_by_year",
        selectable=select(
            [
                PlayerId.db_player_id.label("id"),
                PitchFx.season_id.label("season_id"),
                PlayerId.mlb_id.label("mlb_id"),
                PlayerId.mlb_name.label("name"),
                func.count(PitchFx.id).label("total_pitches"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CH", 1)], else_=0)).label("ch_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "CU", 1)], else_=0)).label("cu_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "EP", 1)], else_=0)).label("ep_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FA", 1)], else_=0)).label("fa_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FC", 1)], else_=0)).label("fc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FF", 1)], else_=0)).label("ff_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FS", 1)], else_=0)).label("fs_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FT", 1)], else_=0)).label("ft_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "FO", 1)], else_=0)).label("fo_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "IN", 1)], else_=0)).label("in_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KC", 1)], else_=0)).label("kc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "KN", 1)], else_=0)).label("kn_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "PO", 1)], else_=0)).label("po_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SC", 1)], else_=0)).label("sc_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SI", 1)], else_=0)).label("si_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "SL", 1)], else_=0)).label("sl_count"),
                func.sum(case([(PitchFx.mlbam_pitch_name == "UN", 1)], else_=0)).label("un_count"),
            ]
        )
        .where(or_(PitchFx.stand == "L", PitchFx.stand == "R"))
        .select_from(
            join(
                PlayerId,
                PitchFx,
                PlayerId.db_player_id == PitchFx.pitcher_id,
                isouter=True,
            )
        )
        .group_by(PitchFx.season_id)
        .group_by(PlayerId.db_player_id)
        .order_by(PlayerId.mlb_name),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )



"""Calculate average time between pitches during at bat, between at bats and between innings."""
from datetime import datetime, timezone

from events import Events

from vigorish.config.database import Season_Game_PitchApp_View as Season_View
from vigorish.tasks.base import Task
from vigorish.util.datetime_util import TIME_ZONE_NEW_YORK
from vigorish.util.numeric_helpers import trim_data_set
from vigorish.util.regex import PFX_TIMESTAMP_REGEX
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id


class CalculateAverageTimeBetweenPitches(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "find_eligible_games_start",
                "find_eligible_games_complete",
                "calculate_pitch_metrics_start",
                "calculate_pitch_metrics_progress",
                "calculate_pitch_metrics_complete",
            )
        )

    def execute(self, trim_data_sets=True):
        self.events.find_eligible_games_start()
        game_ids = Season_View.get_all_bbref_game_ids_combined_no_missing_pfx(self.db_engine)
        if not game_ids:
            return Result.Fail("No games meet the requirements for this process.")
        self.events.find_eligible_games_complete(game_ids)
        self.events.calculate_pitch_metrics_start()
        pitch_samples = []
        at_bat_samples = []
        inning_samples = []
        for num, game_id in enumerate(game_ids, start=1):
            combined_data = self.scraped_data.get_combined_game_data(game_id)
            if not combined_data:
                continue
            result = self.calc_pitch_metrics(combined_data)
            pitch_samples.extend(result[0])
            at_bat_samples.extend(result[2])
            inning_samples.extend(result[4])
            self.events.calculate_pitch_metrics_progress(num)
        self.events.calculate_pitch_metrics_complete()
        metrics = {
            "time_between_pitches": self.process_data_set(
                pitch_samples, trim=trim_data_sets, st_dev=0.2
            ),
            "time_between_at_bats": self.process_data_set(
                at_bat_samples, trim=trim_data_sets, st_dev=0.25
            ),
            "time_between_innings": self.process_data_set(
                inning_samples, trim=trim_data_sets, st_dev=0.007
            ),
        }
        return Result.Ok(metrics)

    def calc_pitch_metrics(self, combined_data):
        pitch_samples = []
        at_bat_samples = []
        inning_samples = []
        prev_ab_pfx = None
        prev_inn_pfx = None
        for half_inning in combined_data["play_by_play_data"]:
            for ab_num, at_bat in enumerate(half_inning["inning_events"], start=1):
                if (
                    at_bat["at_bat_pitchfx_audit"]["pitchfx_error"]
                    or at_bat["at_bat_pitchfx_audit"]["missing_pitchfx_count"]
                ):
                    continue
                for count, pfx in enumerate(at_bat["pitchfx"], start=1):
                    if count == 1 and prev_inn_pfx and ab_num == 1:
                        between_innings = self.get_seconds_between_pitches(prev_inn_pfx, pfx)
                        if between_innings > 0:
                            inning_samples.append(between_innings)
                        prev_inn_pfx = None
                    if count == 1 and prev_ab_pfx and ab_num != 1:
                        between_at_bats = self.get_seconds_between_pitches(prev_ab_pfx, pfx)
                        if between_at_bats > 0:
                            at_bat_samples.append(between_at_bats)
                        prev_ab_pfx = None
                    if count == len(at_bat["pitchfx"]):
                        if ab_num == len(half_inning["inning_events"]):
                            prev_inn_pfx = pfx
                        else:
                            prev_ab_pfx = pfx
                        continue
                    between_pitches = self.get_seconds_between_pitches(
                        pfx, at_bat["pitchfx"][count]
                    )
                    if between_pitches > 0:
                        pitch_samples.append(between_pitches)
        return (
            pitch_samples,
            self.process_data_set(pitch_samples),
            at_bat_samples,
            self.process_data_set(at_bat_samples),
            inning_samples,
            self.process_data_set(inning_samples),
        )

    def get_seconds_between_pitches(self, pitch1, pitch2):
        pitch1_thrown = self.get_time_pitch_thrown(pitch1)
        pitch2_thrown = self.get_time_pitch_thrown(pitch2)
        if not pitch1_thrown or not pitch2_thrown:
            return 0
        return (pitch2_thrown - pitch1_thrown).total_seconds()

    def get_time_pitch_thrown(self, pfx):
        match = PFX_TIMESTAMP_REGEX.match(pfx["park_sv_id"])
        if not match:
            return None
        group_dict = match.groupdict()
        game_dict = validate_bbref_game_id(pfx["bbref_game_id"]).value
        try:
            timestamp = datetime(
                game_dict["game_date"].year,
                int(group_dict["month"]),
                int(group_dict["day"]),
                int(group_dict["hour"]),
                int(group_dict["minute"]),
                int(group_dict["second"]),
            )
        except ValueError:
            return None
        return timestamp.replace(tzinfo=timezone.utc).astimezone(TIME_ZONE_NEW_YORK)

    def process_data_set(self, data_set, trim=False, st_dev=1):
        if not data_set or not isinstance(data_set, list):
            return {"error": "Data set is empty or is not a valid list"}
        if trim:
            data_set = trim_data_set(data_set, st_dev_limit=st_dev)
        if not data_set:
            return {"error": f"Trim process with st_dev_limit={st_dev} eliminated all data"}
        results = {
            "total": int(sum(data_set)),
            "count": int(len(data_set)),
            "avg": round(sum(data_set) / len(data_set), 1),
            "max": int(max(data_set)),
            "min": int(min(data_set)),
            "range": int(max(data_set) - min(data_set)),
            "trim": trim,
        }
        if trim:
            results["trim_stdev"] = st_dev
        return results



from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, dateformat
from dateutil import tz
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from vigorish.config.database import Base
from vigorish.models.season import Season
from vigorish.util.dataclass_helpers import dict_from_dataclass, sanitize_row_dict
from vigorish.util.dt_format_strings import (
    DATE_ONLY,
    DATE_ONLY_2,
    DATE_ONLY_TABLE_ID,
    DT_STR_FORMAT,
)
from vigorish.util.list_helpers import display_dict
from vigorish.util.string_helpers import get_brooks_team_id, validate_brooks_game_id


class GameScrapeStatus(Base):

    __tablename__ = "scrape_status_game"
    id = Column(Integer, primary_key=True)
    game_date = Column(DateTime)
    game_time_hour = Column(Integer)
    game_time_minute = Column(Integer)
    game_time_zone = Column(String)
    bbref_game_id = Column(String)
    bb_game_id = Column(String)
    scraped_bbref_boxscore = Column(Integer, default=0)
    scraped_brooks_pitch_logs = Column(Integer, default=0)
    combined_data_success = Column(Integer, default=0)
    combined_data_fail = Column(Integer, default=0)
    pitch_app_count_bbref = Column(Integer, default=0)
    pitch_app_count_brooks = Column(Integer, default=0)
    total_pitch_count_bbref = Column(Integer, default=0)
    scrape_status_date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    scrape_status_pitchfx = relationship("PitchAppScrapeStatus", backref="scrape_status_game")
    pitch_app_status = relationship(
        "Game_PitchApp_View",
        backref="original",
        uselist=False,
        primaryjoin="GameScrapeStatus.id==Game_PitchApp_View.id",
        foreign_keys="Game_PitchApp_View.id",
    )

    @hybrid_property
    def game_start_time(self):
        if not self.game_time_hour:
            return None
        game_start_time = datetime(
            year=self.game_date.year,
            month=self.game_date.month,
            day=self.game_date.day,
            hour=self.game_time_hour,
            minute=self.game_time_minute,
        )
        return game_start_time.replace(tzinfo=tz.gettz(self.game_time_zone))

    @hybrid_property
    def game_date_time_str(self):
        return (
            self.game_start_time.strftime(DT_STR_FORMAT)
            if self.game_start_time
            else self.game_date.strftime(DATE_ONLY_2)
            if self.game_date
            else None
        )

    @hybrid_property
    def away_team_id_bb(self):
        game_dict = validate_brooks_game_id(self.bb_game_id).value
        return game_dict["away_team_id"]

    @hybrid_property
    def home_team_id_bb(self):
        game_dict = validate_brooks_game_id(self.bb_game_id).value
        return game_dict["home_team_id"]

    @hybrid_property
    def pitch_app_ids(self):
        return [pfx.pitch_app_id for pfx in self.scrape_status_pitchfx]

    @hybrid_property
    def pitch_app_ids_no_pitchfx_data(self):
        return [pfx.pitch_app_id for pfx in self.scrape_status_pitchfx if pfx.no_pitchfx_data == 1]

    @hybrid_property
    def pitch_app_count_pitchfx(self):
        return self.pitch_app_status.total_pitchfx if self.pitch_app_status.total_pitchfx else 0

    @hybrid_property
    def total_pitch_apps_scraped_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx_scraped
            if self.pitch_app_status.total_pitchfx_scraped
            else 0
        )

    @hybrid_property
    def total_pitch_apps_no_pitchfx_data(self):
        return (
            self.pitch_app_status.total_no_pitchfx_data
            if self.pitch_app_status.total_no_pitchfx_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_with_pitchfx_data(self):
        return self.pitch_app_count_pitchfx - self.total_pitch_apps_no_pitchfx_data

    @hybrid_property
    def total_pitch_apps_combined_data(self):
        return (
            self.pitch_app_status.total_combined_pitchfx_bbref_data
            if self.pitch_app_status.total_combined_pitchfx_bbref_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_contains_patched_data(self):
        return len(pfx for pfx in self.scrape_status_pitchfx if pfx.contains_patched_data)

    @hybrid_property
    def total_pitch_apps_pitchfx_error(self):
        total_extra_pfx = self.pitch_app_status.total_pitchfx_error
        return total_extra_pfx if total_extra_pfx else 0

    @hybrid_property
    def total_pitch_apps_invalid_pitchfx(self):
        total_invalid_pfx = self.pitch_app_status.total_invalid_pitchfx
        return total_invalid_pfx if total_invalid_pfx else 0

    @hybrid_property
    def total_pitch_apps_pitchfx_is_valid(self):
        return sum(
            not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx
        )

    @hybrid_property
    def pitch_app_ids_with_pitchfx_data_not_only_patched(self):
        return [
            pfx.pitch_app_id
            for pfx in self.scrape_status_pitchfx
            if pfx.combined_pitchfx_bbref_data
            and not pfx.no_pitchfx_data
            and not pfx.contains_only_patched_data
        ]

    @hybrid_property
    def total_pitch_count_pitch_logs(self):
        return (
            self.pitch_app_status.total_pitch_count_pitch_log
            if self.pitch_app_status.total_pitch_count_pitch_log
            else 0
        )

    @hybrid_property
    def total_pitch_count_bbref_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_bbref
            if self.pitch_app_status.total_pitch_count_bbref
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx
            if self.pitch_app_status.total_pitch_count_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx_audited
            if self.pitch_app_status.total_pitch_count_pitchfx_audited
            else 0
        )

    @hybrid_property
    def total_duplicate_pitchfx_removed_count(self):
        return (
            self.pitch_app_status.total_duplicate_pitchfx_removed_count
            if self.pitch_app_status.total_duplicate_pitchfx_removed_count
            else 0
        )

    @hybrid_property
    def total_patched_pitchfx_count(self):
        return (
            self.pitch_app_status.total_patched_pitchfx_count
            if self.pitch_app_status.total_patched_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_missing_pitchfx_count(self):
        return (
            self.pitch_app_status.total_missing_pitchfx_count
            if self.pitch_app_status.total_missing_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_extra_pitchfx_count(self):
        return (
            self.pitch_app_status.total_extra_pitchfx_count
            if self.pitch_app_status.total_extra_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_extra_pitchfx_removed_count(self):
        return (
            self.pitch_app_status.total_extra_pitchfx_removed_count
            if self.pitch_app_status.total_extra_pitchfx_removed_count
            else 0
        )

    @hybrid_property
    def total_batters_faced_bbref(self):
        return (
            self.pitch_app_status.total_batters_faced_bbref
            if self.pitch_app_status.total_batters_faced_bbref
            else 0
        )

    @hybrid_property
    def total_batters_faced_pitchfx(self):
        return (
            self.pitch_app_status.total_batters_faced_pitchfx
            if self.pitch_app_status.total_batters_faced_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_complete(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_complete
            if self.pitch_app_status.total_at_bats_pitchfx_complete
            else 0
        )

    @hybrid_property
    def total_at_bats_patched_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_patched_pitchfx
            if self.pitch_app_status.total_at_bats_patched_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_missing_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_missing_pitchfx
            if self.pitch_app_status.total_at_bats_missing_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_extra_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_extra_pitchfx
            if self.pitch_app_status.total_at_bats_extra_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_extra_pitchfx_removed(self):
        return (
            self.pitch_app_status.total_at_bats_extra_pitchfx_removed
            if self.pitch_app_status.total_at_bats_extra_pitchfx_removed
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_error(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_error
            if self.pitch_app_status.total_at_bats_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_at_bats_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_invalid_pitchfx
            if self.pitch_app_status.total_at_bats_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def no_pitchfx_data_for_all_pitch_apps(self):
        if not self.scraped_brooks_pitch_logs:
            return False
        return all(pfx.no_pitchfx_data for pfx in self.scrape_status_pitchfx)

    @hybrid_property
    def no_pitchfx_data_for_any_pitch_apps(self):
        if not self.scraped_brooks_pitch_logs:
            return False
        return any(pfx.no_pitchfx_data for pfx in self.scrape_status_pitchfx)

    @hybrid_property
    def scraped_all_pitchfx_logs(self):
        if not self.scraped_brooks_pitch_logs:
            return False
        if not self.scrape_status_pitchfx:
            return True
        return self.pitch_app_count_pitchfx == self.total_pitch_apps_scraped_pitchfx

    @hybrid_property
    def combined_data_for_all_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return False
        if not self.scrape_status_pitchfx:
            return False
        return self.pitch_app_count_pitchfx == self.total_pitch_apps_combined_data

    @hybrid_property
    def pitchfx_error_for_any_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return False
        if not self.scrape_status_pitchfx:
            return False
        return any((pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx)

    @hybrid_property
    def pitchfx_is_valid_for_all_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return True
        if not self.scrape_status_pitchfx:
            return True
        return all(
            not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx
        )

    def __repr__(self):
        return f"<GameScrapeStatus bbref_game_id={self.bbref_game_id}>"

    def as_dict(self):
        d = {}
        d["bbref_game_id"] = self.bbref_game_id
        d["bb_game_id"] = self.bb_game_id
        d["game_date_time_str"] = self.game_date_time_str
        d["scraped_bbref_boxscore"] = self.scraped_bbref_boxscore
        d["scraped_brooks_pitch_logs"] = self.scraped_brooks_pitch_logs
        d["scraped_all_pitchfx_logs"] = self.scraped_all_pitchfx_logs
        d["combined_data_success"] = self.combined_data_success
        d["combined_data_fail"] = self.combined_data_fail
        d["pitch_app_count_bbref"] = self.pitch_app_count_bbref
        d["pitch_app_count_brooks"] = self.pitch_app_count_brooks
        d["pitch_app_count_pitchfx"] = self.pitch_app_count_pitchfx
        d["total_pitch_apps_scraped_pitchfx"] = self.total_pitch_apps_scraped_pitchfx
        d["total_pitch_apps_no_pitchfx_data"] = self.total_pitch_apps_no_pitchfx_data
        d["total_pitch_apps_with_pitchfx_data"] = self.total_pitch_apps_with_pitchfx_data
        d["total_pitch_apps_combined_data"] = self.total_pitch_apps_combined_data
        d["total_pitch_apps_pitchfx_error"] = self.total_pitch_apps_pitchfx_error
        d["total_pitch_apps_invalid_pitchfx"] = self.total_pitch_apps_invalid_pitchfx
        d["total_pitch_apps_pitchfx_is_valid"] = self.total_pitch_apps_pitchfx_is_valid
        d["total_pitch_count_pitch_logs"] = self.total_pitch_count_pitch_logs
        d["total_pitch_count_bbref"] = self.total_pitch_count_bbref
        d["total_pitch_count_bbref_audited"] = self.total_pitch_count_bbref_audited
        d["total_pitch_count_pitchfx"] = self.total_pitch_count_pitchfx
        d["total_pitch_count_pitchfx_audited"] = self.total_pitch_count_pitchfx_audited
        d["total_duplicate_pitchfx_removed_count"] = self.total_duplicate_pitchfx_removed_count
        d["total_missing_pitchfx_count"] = self.total_missing_pitchfx_count
        d["total_extra_pitchfx_count"] = self.total_extra_pitchfx_count
        d["total_extra_pitchfx_removed_count"] = self.total_extra_pitchfx_removed_count
        d["total_batters_faced_bbref"] = self.total_batters_faced_bbref
        d["total_batters_faced_pitchfx"] = self.total_batters_faced_pitchfx
        d["total_at_bats_pitchfx_complete"] = self.total_at_bats_pitchfx_complete
        d["total_at_bats_missing_pitchfx"] = self.total_at_bats_missing_pitchfx
        d["total_at_bats_extra_pitchfx"] = self.total_at_bats_extra_pitchfx
        d["total_at_bats_extra_pitchfx_removed"] = self.total_at_bats_extra_pitchfx_removed
        d["total_at_bats_pitchfx_error"] = self.total_at_bats_pitchfx_error
        d["total_at_bats_invalid_pitchfx"] = self.total_at_bats_invalid_pitchfx
        d["scraped_all_pitchfx_logs"] = self.scraped_all_pitchfx_logs
        d["combined_data_for_all_pitchfx_logs"] = self.combined_data_for_all_pitchfx_logs
        d["pitchfx_error_for_any_pitchfx_logs"] = self.pitchfx_error_for_any_pitchfx_logs
        d["pitchfx_is_valid_for_all_pitchfx_logs"] = self.pitchfx_is_valid_for_all_pitchfx_logs
        return d

    def display(self):
        game_status_dict = self.as_dict()
        title = f"SCRAPE STATUS FOR GAME: {self.bbref_game_id}"
        display_dict(game_status_dict, title=title)

    def status_report(self):
        bbref_game_id = self.bbref_game_id if self.bbref_game_id else "BBREF_GAME_ID IS MISSING!"
        bb_game_id = self.bb_game_id if self.bb_game_id else "BB_GAME_ID IS MISSING!"
        scraped_bbref_boxscore = "YES" if self.scraped_bbref_boxscore == 1 else "NO"
        scraped_brooks_pitch_logs = "YES" if self.scraped_brooks_pitch_logs == 1 else "NO"
        scraped_all_pitchfx_logs = "YES" if self.scraped_all_pitchfx_logs else "NO"
        combined_data_for_all_pitchfx_logs = (
            "YES" if self.combined_data_for_all_pitchfx_logs else "NO"
        )
        pitchfx_error_for_any_pitchfx_logs = (
            "YES" if self.pitchfx_error_for_any_pitchfx_logs else "NO"
        )
        total_pitchfx_removed_count = (
            self.total_duplicate_pitchfx_removed_count + self.total_extra_pitchfx_removed_count
        )
        return (
            f"BBRef Game ID................................: {bbref_game_id}\n"
            f"brooksbaseball.net Game ID...................: {bb_game_id}\n"
            f"Game Date-Time...............................: {self.game_date_time_str}\n"
            f"Scraped BBRef Boxscore.......................: {scraped_bbref_boxscore}\n"
            f"Scraped Brooks Pitch Logs....................: {scraped_brooks_pitch_logs}\n"
            f"PitchFx Logs Scraped.........................: {scraped_all_pitchfx_logs} "
            f"{self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_pitchfx}\n"
            "Combined BBRef/PitchFX Data..................: "
            f"{combined_data_for_all_pitchfx_logs}\n"
            f"PitchFX Data Errors (Valid AB/Invalid AB)....: {pitchfx_error_for_any_pitchfx_logs} "
            f"{self.total_pitch_apps_pitchfx_error}/{self.total_pitch_apps_invalid_pitchfx}\n"
            f"Pitch App Count (BBRef/Brooks)...............: "
            f"{self.pitch_app_count_bbref}/{self.pitch_app_count_brooks}\n"
            f"Pitch App Count (PFx/data/no data)...........: {self.pitch_app_count_pitchfx}/"
            f"{self.total_pitch_apps_with_pitchfx_data}/{self.total_pitch_apps_no_pitchfx_data}\n"
            f"Pitch Count (BBRef/Brooks/PFx)...............: {self.total_pitch_count_bbref}/"
            f"{self.total_pitch_count_pitch_logs}/{self.total_pitch_count_pitchfx}\n"
            "Pitch Count Audited (BBRef/PFx/Removed)......: "
            f"{self.total_pitch_count_bbref_audited}/{self.total_pitch_count_pitchfx_audited}/"
            f"{total_pitchfx_removed_count}\n"
        )

    def as_csv_dict(self):
        return dict_from_dataclass(self, GameScrapeStatusCsvRow, date_format=DATE_ONLY)

    @classmethod
    def get_csv_col_names(cls):
        return [name for name in GameScrapeStatusCsvRow.__dataclass_fields__.keys()]

    @classmethod
    def export_table_as_csv(cls, db_session):
        col_names = ",".join(cls.get_csv_col_names())
        csv_dicts = (obj.as_csv_dict() for obj in db_session.query(cls).all())
        csv_rows = (",".join(sanitize_row_dict(d, date_format=DATE_ONLY)) for d in csv_dicts)
        yield col_names
        for row in csv_rows:
            yield row

    @classmethod
    def find_by_bbref_game_id(cls, db_session, bbref_game_id):
        return db_session.query(cls).filter_by(bbref_game_id=bbref_game_id).first()

    @classmethod
    def find_by_bb_game_id(cls, db_session, bb_game_id):
        return db_session.query(cls).filter_by(bb_game_id=bb_game_id).first()

    @classmethod
    def get_all_scraped_bbref_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bbref_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_bbref_boxscore == 1
        ]

    @classmethod
    def get_all_unscraped_bbref_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bbref_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_bbref_boxscore == 0
        ]

    @classmethod
    def get_all_scraped_brooks_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bb_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_brooks_pitch_logs == 1
        ]

    @classmethod
    def get_all_unscraped_brooks_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bb_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_brooks_pitch_logs == 0
        ]

    @classmethod
    def get_all_bbref_game_ids(cls, db_session, season_id):
        return [
            game_status.bbref_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
        ]

    @classmethod
    def get_game_id_dict(cls, db_session, season_id):
        return {
            game_status.bbref_game_id: game_status.id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
        }

    @classmethod
    def get_all_bbref_game_ids_for_date(cls, db_session, date):
        date_id = date.strftime(DATE_ONLY_TABLE_ID)
        return [
            game.bbref_game_id
            for game in db_session.query(cls).filter_by(scrape_status_date_id=int(date_id)).all()
        ]

    @classmethod
    def get_all_brooks_game_ids_for_date(cls, db_session, date):
        date_id = date.strftime(DATE_ONLY_TABLE_ID)
        return [
            game.bb_game_id
            for game in db_session.query(cls)
            .filter_by(scrape_status_date_id=int(date_id))
            .order_by(GameScrapeStatus.scrape_status_date_id)
            .all()
        ]

    @classmethod
    def get_all_bbref_game_ids_for_team(cls, db_session, team_id_br, year):
        season = Season.find_by_year(db_session, year)
        if not season:
            return None
        games_for_season = [
            game for game in db_session.query(cls).filter_by(season_id=season.id).all()
        ]
        games_for_season.sort(key=lambda x: x.scrape_status_date_id)
        bb_team_id = get_brooks_team_id(team_id_br)
        return [
            game.bbref_game_id
            for game in games_for_season
            if game.away_team_id_bb == bb_team_id or game.home_team_id_bb == bb_team_id
        ]

    @classmethod
    def get_game_id_map(cls, db_session):
        all_games = db_session.query(cls).all()
        return {g.bbref_game_id: g.id for g in all_games}


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class GameScrapeStatusCsvRow:
    id: int
    game_date: datetime = None
    game_time_hour: int = 0
    game_time_minute: int = 0
    game_time_zone: str = None
    bbref_game_id: str = None
    bb_game_id: str = None
    scraped_bbref_boxscore: int = 0
    scraped_brooks_pitch_logs: int = 0
    combined_data_success: int = 0
    combined_data_fail: int = 0
    pitch_app_count_bbref: int = 0
    pitch_app_count_brooks: int = 0
    total_pitch_count_bbref: int = 0



"""Populate player_id and player tables with initial data."""
from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, DataclassReader, dateformat
from tqdm import tqdm

from vigorish.config.database import Player, PlayerId
from vigorish.tasks.update_player_maps import UpdatePlayerIdMap
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class PlayerCsvRow:
    playerID: str = None
    birthYear: int = None
    birthMonth: int = None
    birthDay: int = None
    birthCountry: str = None
    birthState: str = None
    birthCity: str = None
    deathYear: int = None
    deathMonth: int = None
    deathDay: int = None
    deathCountry: str = None
    deathState: str = None
    deathCity: str = None
    nameFirst: str = None
    nameLast: str = None
    nameGiven: str = None
    weight: int = None
    height: int = None
    bats: str = None
    throws: str = None
    debut: datetime = None
    finalGame: datetime = None
    retroID: str = None
    bbrefID: str = None


PLAYER_ID_MAP_FILENAME = "bbref_player_id_map.csv"
PEOPLE_CSV_FILENAME = "People.csv"


def populate_players(app, csv_folder):
    db_session = app["db_session"]
    """Populate player_id and player tables with initial data."""
    result = import_idmap_csv(db_session, app, csv_folder)
    if result.failure:
        return result
    db_session.commit()
    result = import_people_csv(db_session, csv_folder)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def import_idmap_csv(db_session, app, csv_folder):
    try:
        update_player_id_map = UpdatePlayerIdMap(app)
        player_id_map_csv = csv_folder.joinpath(PLAYER_ID_MAP_FILENAME)
        player_id_map = update_player_id_map.read_bbref_player_id_map_from_file(player_id_map_csv)
        with tqdm(
            total=len(player_id_map),
            desc="Populating player_id table...",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ) as pbar:
            for id_map in player_id_map:
                db_session.add(
                    PlayerId(
                        mlb_id=int(id_map.mlb_ID),
                        mlb_name=id_map.name_common,
                        bbref_id=id_map.player_ID,
                        bbref_name=None,
                    )
                )
                pbar.update()
        return Result.Ok()
    except Exception as e:
        error = "Error: {error}".format(error=repr(e))
        db_session.rollback()
        return Result.Fail(error)


def import_people_csv(db_session, csv_folder):
    player_csv_file = csv_folder.joinpath(PEOPLE_CSV_FILENAME)
    csv_text = player_csv_file.read_text()
    total_rows = len([row for row in csv_text.split("\n") if row])
    try:
        with open(player_csv_file) as player_csv:
            reader = DataclassReader(player_csv, PlayerCsvRow)
            with tqdm(
                total=total_rows,
                desc="Populating player table......",
                unit="row",
                mininterval=0.12,
                maxinterval=5,
                unit_scale=True,
                ncols=90,
            ) as pbar:
                for row in reader:
                    if not (row.birthYear or row.birthMonth or row.birthDay or row.debut):
                        pbar.update()
                        continue
                    player_id = PlayerId.find_by_bbref_id(db_session, row.bbrefID)
                    if not player_id:
                        pbar.update()
                        continue
                    p = Player(
                        mlb_id=player_id.mlb_id,
                        bbref_id=row.bbrefID,
                        name_first=row.nameFirst,
                        name_last=row.nameLast,
                        name_given=row.nameGiven,
                        bats=row.bats,
                        throws=row.throws,
                        weight=row.weight,
                        height=row.height,
                        debut=row.debut,
                        birth_year=row.birthYear,
                        birth_month=row.birthMonth,
                        birth_day=row.birthDay,
                        birth_country=row.birthCountry,
                        birth_state=row.birthState,
                        birth_city=row.birthCity,
                        missing_mlb_id=False,
                    )
                    db_session.add(p)
                    player_id.db_player_id = p.id
                    pbar.update()
        return Result.Ok()
    except Exception as e:
        error = "Error: {error}".format(error=repr(e))
        db_session.rollback()
        return Result.Fail(error)



"""Menu item that combines scraped boxscore data and pitchfx data for a single game."""
import logging
import subprocess
import time
from collections import defaultdict

import enlighten
from getch import pause
from halo import Halo
from tabulate import tabulate

from vigorish.cli.components import (
    audit_report_season_prompt,
    get_random_cli_color,
    get_random_dots_spinner,
    print_message,
    user_options_prompt,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.config.database import Season
from vigorish.constants import EMOJI_DICT, MENU_NUMBERS
from vigorish.enums import AuditError, DataSet, ScrapeCondition
from vigorish.tasks.combine_scraped_data import CombineScrapedDataTask
from vigorish.util.dt_format_strings import DATE_MONTH_NAME
from vigorish.util.list_helpers import flatten_list2d
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id, validate_pitch_app_id

STATUS_BAR_FORMAT = (
    "{total_combined}/{total_games} Games Combined"
    "{fill}"
    "Combining Data for Games Played On: {date_str}"
    "{fill}"
    "Elapsed: {elapsed}"
)
STATUS_BAR_COLOR = "bold_gray100_on_darkviolet"
DATE_BAR_FORMAT = "{desc}{desc_pad}{percentage:3.0f}% |{bar}| {count:{len_total}d}/{total:d} {unit}"

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class CombineScrapedData(MenuItem):
    def __init__(self, app, audit_report):
        super().__init__(app)
        self.combine_data = CombineScrapedDataTask(app)
        self._season = None
        self._date_game_id_map = {}
        self.pbar_manager = enlighten.get_manager()
        self.audit_report = audit_report
        self.menu_item_text = "Combine Game Data"
        self.menu_item_emoji = EMOJI_DICT.get("BANG", "")
        self.exit_menu = False
        self.combine_condition = self.config.get_current_setting(
            "SCRAPED_DATA_COMBINE_CONDITION", DataSet.ALL
        )
        self.audit_type = None
        self.scrape_year = None
        self.current_game_date = None
        self.current_game_id = None
        self.combine_data_success_game_ids = []
        self.combine_data_fail_results = {}
        self.all_pfx_errors = defaultdict(dict)
        self.status_bar = None
        self.date_progress_bar = None
        self.game_progress_bar_success = None
        self.game_progress_bar_fail = None
        self.game_progress_bar_error = None

    @property
    def terminal(self):
        return self.pbar_manager.term if self.pbar_manager else None

    @property
    def game_bar_format(self):
        SUCCESS = EMOJI_DICT.get("PASSED", "")
        FAIL = EMOJI_DICT.get("FAILED", "")
        ERROR = EMOJI_DICT.get("CONFUSED", "")
        return (
            "{desc}{desc_pad}{percentage:3.0f}% |{bar}| "
            + self.terminal.green3(f"{SUCCESS}" + " {count_0:{len_total}d}")
            + " | "
            + self.terminal.red2(f"{FAIL}" + " {count_1:{len_total}d}")
            + " | "
            + self.terminal.yellow2(f"{ERROR}" + " {count_2:{len_total}d}")
            + " | ETA: {eta}"
        )

    @property
    def season(self):
        if not self.scrape_year:
            return None
        if self._season:
            return self._season
        self._season = Season.find_by_year(self.db_session, self.scrape_year)
        return self._season

    @property
    def date_game_id_map(self):
        if not self.scrape_year:
            return {}
        if self._date_game_id_map:
            return self._date_game_id_map
        scraped_game_date_map = self.get_scraped_game_date_map(self.scrape_year)
        self._date_game_id_map = defaultdict(list)
        for game_map in scraped_game_date_map:
            self._date_game_id_map[game_map["game_date"]].append(game_map["game_id"])
        return self._date_game_id_map

    @property
    def all_dates_in_season(self):
        return (
            [game_date.date() for game_date in self.season.get_date_range()]
            if self.season
            else None
        )

    @property
    def all_dates_with_eligible_games(self):
        return sorted(list({game_date for game_date in self.date_game_id_map.keys()}))

    @property
    def total_dates(self):
        return len(self.all_dates_in_season) if self.audit_type == "SEASON" else 1

    @property
    def all_eligible_games_in_season(self):
        return sorted(flatten_list2d([game_ids for game_ids in self.date_game_id_map.values()]))

    @property
    def total_games(self):
        return (
            sum(len(game_ids) for game_ids in self.date_game_id_map.values())
            if self.audit_type == "SEASON"
            else len(self.date_game_id_map.get(self.current_game_date, None))
        )

    @property
    def total_combined_success(self):
        return len(self.combine_data_success_game_ids)

    @property
    def total_combined_fail(self):
        return len(self.combine_data_fail_results)

    @property
    def total_combined(self):
        return (
            self.total_combined_success
            + self.total_combined_fail
            + self.total_games_invalid_pitchfx
            + self.total_games_pitchfx_error
        )

    @property
    def pfx_errors(self):
        return self.all_pfx_errors[AuditError.PITCHFX_ERROR]

    @property
    def invalid_pfx(self):
        return self.all_pfx_errors[AuditError.INVALID_PITCHFX_DATA]

    @property
    def total_games_pitchfx_error(self):
        return len(self.pfx_errors)

    @property
    def total_games_invalid_pitchfx(self):
        return len(self.invalid_pfx)

    @property
    def total_games_any_pfx_error(self):
        return len(self.pfx_errors) + len(self.invalid_pfx)

    @property
    def game_ids_any_pitchfx_error(self):
        unique_game_ids = {
            game_id for error_dict in self.all_pfx_errors.values() for game_id in error_dict.keys()
        }
        return list(unique_game_ids)

    @property
    def pitch_apps_pitchfx_error(self):
        return flatten_list2d(
            list(pitch_app_dict.keys()) for pitch_app_dict in self.pfx_errors.values()
        )

    @property
    def pitch_apps_invalid_pitchfx(self):
        return flatten_list2d(
            list(pitch_app_dict.keys()) for pitch_app_dict in self.invalid_pfx.values()
        )

    @property
    def pitch_apps_any_pfx_error(self):
        return sorted(list(set(self.pitch_apps_pitchfx_error + self.pitch_apps_invalid_pitchfx)))

    @property
    def total_pitch_apps_pitchfx_error(self):
        return len(self.pitch_apps_pitchfx_error)

    @property
    def total_pitch_apps_invalid_pitchfx(self):
        return len(self.pitch_apps_invalid_pitchfx)

    @property
    def total_pitch_apps_any_pitchfx_error(self):
        return len(self.pitch_apps_pitchfx_error) + len(self.pitch_apps_invalid_pitchfx)

    @property
    def game_id_pitch_app_id_map(self):
        id_map = defaultdict(list)
        for pitch_app_id in self.pitch_apps_any_pfx_error:
            result = validate_pitch_app_id(pitch_app_id)
            if result.failure:
                return {}
            pitch_app_dict = result.value
            id_map[pitch_app_dict["game_id"]].append(pitch_app_id)
        return id_map

    @property
    def at_bats_pitchfx_error(self):
        return flatten_list2d(
            pitch_app_dict.values() for pitch_app_dict in self.all_pfx_errors.values()
        )

    @property
    def at_bats_invalid_pitchfx(self):
        return flatten_list2d(
            pitch_app_dict.values() for pitch_app_dict in self.invalid_pfx.values()
        )

    @property
    def total_at_bats_pitchfx_error(self):
        return len(self.at_bats_pitchfx_error)

    @property
    def total_at_bats_invalid_pitchfx(self):
        return len(self.at_bats_invalid_pitchfx)

    @property
    def total_at_bats_any_pitchfx_error(self):
        return len(self.at_bats_pitchfx_error) + len(self.at_bats_invalid_pitchfx)

    @property
    def failed_game_ids(self):
        return list(game_id for game_id in self.combine_data_fail_results.keys())

    @property
    def combined_success_and_no_pfx_errors(self):
        return not self.total_combined_fail and not self.total_games_any_pfx_error

    def launch(self):
        subprocess.run(["clear"])
        exit_menu = False
        while not exit_menu:
            result = self.audit_type_prompt()
            if result.failure:
                exit_menu = True
                continue
            self.audit_type = result.value
            if self.audit_type == "SEASON":
                result = self.combine_games_for_season()
            if self.audit_type == "DATE":
                result = self.combine_games_for_date()
            if self.audit_type == "GAME":
                result = self.combine_single_game()
                if result.success:
                    exit_menu = True
                    continue
            if result.failure:
                continue
            self.display_results()
        return Result.Ok(self.exit_menu)

    def combine_games_for_season(self):
        result = audit_report_season_prompt(self.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        self.pbar_manager = enlighten.get_manager()
        self.init_progress_bars(game_date=self.all_dates_in_season[0])
        subprocess.run(["clear"])
        for game_date in self.all_dates_in_season:
            if self.every_eligible_game_is_combined():
                num_days_remaining = self.get_number_of_days_remaining()
                self.update_progress_bars(game_date)
                self.date_progress_bar.update(num_days_remaining)
                LOGGER.info(f"Processed all eligible games for MLB {self.scrape_year}.")
                time.sleep(1.5)
                break
            game_ids = self.date_game_id_map.get(game_date, None)
            if not game_ids:
                self.update_progress_bars(game_date)
                self.date_progress_bar.update()
                time.sleep(0.75)
                continue
            result = self.combine_selected_games(game_date, game_ids)
            if result.failure:
                return result
            self.date_progress_bar.update()
        self.close_progress_bars()
        return Result.Ok()

    def init_progress_bars(self, game_date):
        date_str = game_date.strftime(DATE_MONTH_NAME)
        self.status_bar = self.pbar_manager.status_bar(
            status_format=STATUS_BAR_FORMAT,
            color=STATUS_BAR_COLOR,
            justify=enlighten.Justify.CENTER,
            total_combined=self.total_combined,
            total_games=self.total_games,
            date_str=date_str,
        )
        self.date_progress_bar = self.pbar_manager.counter(
            total=self.total_dates,
            bar_format=DATE_BAR_FORMAT,
            desc=date_str,
            unit="days",
            color="cyan2",
            autorefresh=True,
        )
        self.game_progress_bar_success = self.pbar_manager.counter(
            total=self.total_games,
            bar_format=self.game_bar_format,
            desc="Combining data for game...",
            unit="games",
            color="green3",
            autorefresh=True,
        )
        self.game_progress_bar_fail = self.game_progress_bar_success.add_subcounter("red2")
        self.game_progress_bar_error = self.game_progress_bar_success.add_subcounter("yellow2")

    def every_eligible_game_is_combined(self):
        return self.total_combined == self.total_games

    def get_number_of_days_remaining(self):
        return self.total_dates - self.date_progress_bar.count

    def combine_selected_games(self, game_date, game_ids):
        for bbref_game_id in game_ids:
            fail_results = []
            self.current_game_id = bbref_game_id
            self.update_progress_bars(game_date)
            # LOGGER.info(f"Begin combining scraped data for game: {bbref_game_id}")
            result = self.combine_data.execute(bbref_game_id)
            if not result["gather_scraped_data_success"]:
                LOGGER.info(f"Unable to combine data for game: {bbref_game_id}")
                LOGGER.info(f"An error occurred gathering scraped data for game: {bbref_game_id}")
                return Result.Fail(result["error"])
            if not result["combined_data_success"]:
                self.combine_data_fail_results[bbref_game_id] = result["error"]
                self.game_progress_bar_fail.update()
                LOGGER.info(f"Failed to combine scraped data for game: {bbref_game_id}")
                LOGGER.info(result["error"])
                continue
            if not result["update_pitch_apps_success"]:
                LOGGER.info(f"Error occurred updating pitch apps for game: {bbref_game_id}")
                LOGGER.info(result["error"])
                return Result.Fail(result["error"])
            pfx_errors = result["results"]["pfx_errors"]
            if pfx_errors.get("pitchfx_error", []):
                self.pfx_errors[bbref_game_id] = pfx_errors["pitchfx_error"]
                fail_results.append(pfx_errors["pitchfx_error"])
            if pfx_errors.get("invalid_pitchfx", []):
                self.invalid_pfx[bbref_game_id] = pfx_errors["invalid_pitchfx"]
                fail_results.append(pfx_errors["invalid_pitchfx"])
            if fail_results:
                self.game_progress_bar_error.update()
                self.log_pfx_data_error_details(bbref_game_id, fail_results)
            else:
                self.combine_data_success_game_ids.append(bbref_game_id)
                self.game_progress_bar_success.update()
                # LOGGER.info(f"Successfully combined scraped data for game: {bbref_game_id}")
        return Result.Ok()

    def update_progress_bars(self, game_date):
        date_str = game_date.strftime(DATE_MONTH_NAME)
        self.status_bar.update(
            total_combined=self.total_combined,
            total_games=self.total_games,
            date_str=date_str,
        )
        self.date_progress_bar.desc = date_str
        self.game_progress_bar_success.desc = self.current_game_id

    def log_pfx_data_error_details(self, bbref_game_id, fail_results):
        total_pitch_apps = sum(len(f.keys()) for f in fail_results)
        pitch_apps_plural = "pitch appearances" if total_pitch_apps > 1 else "pitch appearance"
        total_at_bats = sum(len(at_bat_ids) for f in fail_results for at_bat_ids in f.values())
        at_bats_plural = "at bats" if total_at_bats > 1 else "at bat"
        LOGGER.info(f"PitchFX data could not be reconciled for game: {bbref_game_id}")
        LOGGER.info(
            f"{total_pitch_apps} {pitch_apps_plural} with data errors ({total_at_bats} "
            f"total {at_bats_plural})\n"
        )

    def close_progress_bars(self):
        self.status_bar.close()
        self.date_progress_bar.close()
        self.game_progress_bar_success.close()
        self.pbar_manager.stop()

    def display_results(self):
        subprocess.run(["clear"])
        if self.combined_success_and_no_pfx_errors:
            plural = "games total" if self.total_games > 1 else "game"
            success_message = f"\nAll game data ({self.total_games} {plural}) combined, no errors"
            print_message(success_message, wrap=False, fg="bright_cyan", bold=True)
        if self.failed_game_ids:
            self.display_games_failed_to_combine()
        if self.all_pfx_errors:
            self.display_pitchfx_errors()
        pause(message="Press any key to continue...")

    def display_games_failed_to_combine(self):
        error_message = (
            f"Error prevented scraped data being combined for {len(self.failed_game_ids)} " "games:"
        )
        error_details = [
            {"bbref_game_id": game_id, "error": error}
            for game_id, error in self.combine_data_fail_results
        ]
        print_message(error_message, wrap=False, fg="bright_red", bold=True)
        print_message(tabulate(error_details, headers="keys"), wrap=False, fg="bright_red")

    def display_pitchfx_errors(self):
        games_plural = "games contain" if self.total_games_any_pfx_error > 1 else "game contains"
        ab_plural = "at bats" if self.total_at_bats_any_pitchfx_error > 1 else "at bat"
        message = (
            f"{self.total_games_any_pfx_error} {games_plural} invalid PitchFX data for a total "
            f"of {len(self.total_at_bats_any_pitchfx_error)} {ab_plural}, you can view details "
            "of each at bat and attempt to fix these errors using the Investigate Failures menu."
        )
        print_message(message, fg="bright_cyan")
        print()

    def combine_games_for_date(self):
        result = audit_report_season_prompt(self.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        result = self.game_date_prompt()
        if result.failure:
            return result
        self.current_game_date = result.value
        self.pbar_manager = enlighten.get_manager()
        self.init_progress_bars(game_date=self.current_game_date)
        subprocess.run(["clear"])
        game_ids = self.date_game_id_map.get(self.current_game_date, None)
        if not game_ids:
            game_date_str = self.current_game_date.strftime(DATE_MONTH_NAME)
            message = f"All games on {game_date_str} have already been combined."
            print_message(message, fg="bright_cyan", bold=True)
            self.close_progress_bars()
            return Result.Ok()
        result = self.combine_selected_games(self.current_game_date, game_ids)
        self.date_progress_bar.update()
        self.close_progress_bars()
        return result

    def combine_single_game(self):
        result = audit_report_season_prompt(self.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        result = self.game_id_prompt()
        if result.failure:
            return result
        combine_game_id = result.value
        return self.combine_scraped_data_for_game(combine_game_id)

    def combine_scraped_data_for_game(self, combine_game_id):
        subprocess.run(["clear"])
        spinner = Halo(color=get_random_cli_color(), spinner=get_random_dots_spinner())
        spinner.text = f"Combining scraped data for {combine_game_id}..."
        spinner.start()
        result = self.combine_data.execute(combine_game_id)
        if not (
            result["gather_scraped_data_success"]
            and result["combined_data_success"]
            and result["update_pitch_apps_success"]
        ):
            spinner.fail(f"Failed to combine data for {combine_game_id}!")
            print_message(result["error"], wrap=False, fg="bright_red", bold=True)
            return Result.Fail(result["error"])
        spinner.stop()
        pfx_errors = result["results"]["pfx_errors"]
        if pfx_errors.get("pitchfx_error", []):
            self.pfx_errors[combine_game_id] = pfx_errors["pitchfx_error"]
        if pfx_errors.get("invalid_pitchfx", []):
            self.invalid_pfx[combine_game_id] = pfx_errors["invalid_pitchfx"]
        if self.total_pitch_apps_any_pitchfx_error > 0:
            pitch_apps_plural = (
                "pitch appearances"
                if self.total_pitch_apps_any_pitchfx_error > 1
                else "pitch appearance"
            )
            at_bats_plural = "at bats" if self.total_at_bats_any_pitchfx_error > 1 else "at bat"
            message = (
                f"PitchFX data could not be reconciled for game: {combine_game_id}\n"
                f"{self.total_pitch_apps_any_pitchfx_error} {pitch_apps_plural} with data errors "
                f"({self.total_at_bats_any_pitchfx_error} total {at_bats_plural})\n"
            )
            print_message(message, fg="bright_yellow", bold=True)
        else:
            message = f"All scraped data for {combine_game_id} was successfully combined!"
            print_message(message, fg="bright_cyan", bold=True)
        pause(message="Press any key to continue...")
        return Result.Ok()

    def audit_type_prompt(self):
        prompt = (
            "For games where all data has been scraped, would you like to combine and audit the "
            "data for all games in the same season, all games on a single date or for a single "
            "game?"
        )
        choices = {
            f"{MENU_NUMBERS.get(1)}  By Season": "SEASON",
            f"{MENU_NUMBERS.get(2)}  By Date": "DATE",
            f"{MENU_NUMBERS.get(3)}  By Game": "GAME",
            f"{EMOJI_DICT.get('BACK')} Return to Main Menu": None,
        }
        return user_options_prompt(choices, prompt)

    def game_date_prompt(self):
        choices = {
            f"{EMOJI_DICT.get('BOLT')} {game_date.strftime(DATE_MONTH_NAME)}": game_date
            for game_date in self.all_dates_with_eligible_games
        }
        choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select a date to combine scraped data for all games that took place on that date"
        return user_options_prompt(choices, prompt)

    def game_id_prompt(self):
        choices = {
            f"{EMOJI_DICT.get('BOLT')} {game_id}": game_id
            for game_id in self.all_eligible_games_in_season
        }
        choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select the BBRef.com Game ID to combine scraped data:"
        return user_options_prompt(choices, prompt)

    def get_scraped_game_date_map(self, year):
        return (
            self.get_never_audited_game_date_map(year)
            if self.combine_condition == ScrapeCondition.ONLY_MISSING_DATA
            else self.get_all_eligible_game_date_map(year)
            if self.combine_condition == ScrapeCondition.ALWAYS
            else None
        )

    def get_never_audited_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.audit_report[year]["scraped"]]

    def get_successful_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.audit_report[year]["successful"]]

    def get_failed_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.audit_report[year]["failed"]]

    def get_all_eligible_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.get_all_eligible_game_ids(year)]

    def get_all_eligible_game_ids(self, year):
        return (
            self.audit_report[year]["scraped"]
            + self.audit_report[year]["successful"]
            + self.audit_report[year]["failed"]
            + self.audit_report[year]["pfx_error"]
            + self.audit_report[year]["invalid_pfx"]
        )

    def get_game_id_date_map(self, bbref_game_id):
        game_dict = validate_bbref_game_id(bbref_game_id).value
        return {"game_id": bbref_game_id, "game_date": game_dict["game_date"], "count": 1}



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, GameScrapeStatus, PitchAppScrapeStatus


class Game_PitchApp_View(Base):
    __table__ = create_view(
        name="game_pitch_app_status",
        selectable=select(
            [
                GameScrapeStatus.id.label("id"),
                GameScrapeStatus.bbref_game_id.label("bbref_game_id"),
                func.count(PitchAppScrapeStatus.id).label("total_pitchfx"),
                func.sum(PitchAppScrapeStatus.scraped_pitchfx).label("total_pitchfx_scraped"),
                func.sum(PitchAppScrapeStatus.no_pitchfx_data).label("total_no_pitchfx_data"),
                func.sum(PitchAppScrapeStatus.combined_pitchfx_bbref_data).label(
                    "total_combined_pitchfx_bbref_data"
                ),
                func.sum(PitchAppScrapeStatus.batters_faced_bbref).label(
                    "total_batters_faced_bbref"
                ),
                func.sum(PitchAppScrapeStatus.batters_faced_pitchfx).label(
                    "total_batters_faced_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.duplicate_guid_removed_count).label(
                    "total_duplicate_pitchfx_removed_count"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_pitch_log).label(
                    "total_pitch_count_pitch_log"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_bbref).label("total_pitch_count_bbref"),
                func.sum(PitchAppScrapeStatus.pitch_count_pitchfx).label(
                    "total_pitch_count_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_pitchfx_audited).label(
                    "total_pitch_count_pitchfx_audited"
                ),
                func.sum(PitchAppScrapeStatus.patched_pitchfx_count).label(
                    "total_patched_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_patched_pitchfx).label(
                    "total_at_bats_patched_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.missing_pitchfx_count).label(
                    "total_missing_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_missing_pitchfx).label(
                    "total_at_bats_missing_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.extra_pitchfx_count).label(
                    "total_extra_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_extra_pitchfx).label(
                    "total_at_bats_extra_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.extra_pitchfx_removed_count).label(
                    "total_extra_pitchfx_removed_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_extra_pitchfx_removed).label(
                    "total_at_bats_extra_pitchfx_removed"
                ),
                func.sum(PitchAppScrapeStatus.pitchfx_error).label("total_pitchfx_error"),
                func.sum(PitchAppScrapeStatus.total_at_bats_pitchfx_error).label(
                    "total_at_bats_pitchfx_error"
                ),
                func.sum(PitchAppScrapeStatus.invalid_pitchfx).label("total_invalid_pitchfx"),
                func.sum(PitchAppScrapeStatus.invalid_pitchfx_count).label(
                    "total_invalid_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_invalid_pitchfx).label(
                    "total_at_bats_invalid_pitchfx"
                ),
            ]
        )
        .select_from(
            join(
                GameScrapeStatus,
                PitchAppScrapeStatus,
                GameScrapeStatus.id == PitchAppScrapeStatus.scrape_status_game_id,
                isouter=True,
            )
        )
        .group_by(GameScrapeStatus.id),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )



from pprint import pformat

from vigorish.cli.components import print_message
from vigorish.config.database import DateScrapeStatus, Season
from vigorish.enums import StatusReport
from vigorish.util.datetime_util import get_date_range
from vigorish.util.dt_format_strings import DATE_MONTH_NAME, DATE_ONLY
from vigorish.util.result import Result


def report_status_single_date(db_session, game_date, report_type):
    season = Season.find_by_year(db_session, game_date.year)
    date_is_valid = Season.is_date_in_season(db_session, game_date).success
    date_str = game_date.strftime(DATE_ONLY)
    if not date_is_valid:
        error = (
            f"'{date_str}' is not within the {season.name}:\n"
            f"season_start_date: {season.start_date_str}\n"
            f"season_end_date: {season.end_date_str}"
        )
        return Result.Fail(error)
    date_status = DateScrapeStatus.find_by_date(db_session, game_date)
    if not date_status:
        error = f"scrape_status_date does not contain an entry for date: {date_str}"
        return Result.Fail(error)
    missing_pitchfx = (
        report_type == StatusReport.DATE_DETAIL_MISSING_PITCHFX
        or report_type == StatusReport.SINGLE_DATE_WITH_GAME_STATUS
    )
    missing_ids_str = ""
    if missing_pitchfx:
        if date_status.scraped_all_pitchfx_logs:
            missing_ids_str = "All PitchFX logs have been scraped"
        elif date_status.scraped_all_brooks_pitch_logs:
            missing_pitch_app_ids = DateScrapeStatus.get_unscraped_pitch_app_ids_for_date(
                db_session, game_date
            )
            missing_ids_str = (
                f"MISSING: {pformat(missing_pitch_app_ids)}"
                if missing_pitch_app_ids
                else "All PitchFX logs have been scraped"
            )
        else:
            missing_ids_str = (
                "Missing IDs cannot be reported until all pitch logs have been scraped."
            )
    date_str = game_date.strftime(DATE_MONTH_NAME)
    print_message(f"\n### OVERALL STATUS FOR {date_str} ###", fg="bright_cyan", bold=True)
    print_message(date_status.status_report(), wrap=False, fg="bright_cyan")
    if report_type == StatusReport.SINGLE_DATE_WITH_GAME_STATUS:
        print_message(
            f"\n### STATUS FOR EACH GAME PLAYED {date_str} ###", fg="bright_green", bold=True
        )
        print_message(date_status.games_status_report(), wrap=False, fg="bright_green")
    if missing_pitchfx:
        print_message(
            f"\n### MISSING PITCHFX LOGS FOR {date_str} ###", fg="bright_magenta", bold=True
        )
        print_message(missing_ids_str, wrap=False, fg="bright_magenta")
    return Result.Ok()


def report_season_status(db_session, year, report_type):
    if report_type == StatusReport.NONE:
        return Result.Ok()
    season = Season.find_by_year(db_session, year)
    if report_type == StatusReport.SEASON_SUMMARY:
        print_message(f"\n### STATUS REPORT FOR {season.name} ###", fg="bright_yellow", bold=True)
        print_message(season.status_report(), wrap=False, fg="bright_yellow")
        return Result.Ok()
    start_date = season.start_date
    end_date = season.end_date
    return report_date_range_status(db_session, start_date, end_date, report_type)


def report_date_range_status(db_session, start_date, end_date, report_type):
    if report_type == StatusReport.NONE:
        return Result.Ok()
    result = construct_date_range_status(db_session, start_date, end_date, report_type)
    if result.failure:
        return result
    status_date_range = result.value
    return display_date_range_status(
        db_session, start_date, end_date, status_date_range, report_type
    )


def construct_date_range_status(db_session, start_date, end_date, report_type):
    show_all = False
    if (
        report_type == StatusReport.DATE_SUMMARY_ALL_DATES
        or report_type == StatusReport.DATE_DETAIL_ALL_DATES
        or report_type == StatusReport.DATE_DETAIL_MISSING_PITCHFX
    ):
        show_all = True
    status_date_range = []
    for game_date in get_date_range(start_date, end_date):
        date_status = DateScrapeStatus.find_by_date(db_session, game_date)
        if not date_status:
            error = (
                "scrape_status_date does not contain an entry for date: "
                f"{game_date.strftime(DATE_ONLY)}"
            )
            return Result.Fail(error)
        if not show_all and date_status.scraped_all_game_data:
            continue
        status_date_range.append(date_status)
    return Result.Ok(status_date_range)


def display_date_range_status(db_session, start_date, end_date, status_date_range, report_type):
    if (
        report_type == StatusReport.DATE_DETAIL_MISSING_DATA
        or report_type == StatusReport.DATE_DETAIL_ALL_DATES
    ):
        return display_detailed_report_for_date_range(db_session, status_date_range, False)
    if report_type == StatusReport.DATE_DETAIL_MISSING_PITCHFX:
        return display_detailed_report_for_date_range(db_session, status_date_range, True)
    return display_summary_report_for_date_range(
        db_session, start_date, end_date, status_date_range
    )


def display_detailed_report_for_date_range(db_session, status_date_range, missing_pitchfx):
    for date_status in status_date_range:
        game_date_str = date_status.game_date.strftime(DATE_MONTH_NAME)
        missing_ids_str = ""
        if missing_pitchfx:
            if date_status.scraped_all_pitchfx_logs:
                missing_ids_str = "All PitchFX logs have been scraped"
            elif date_status.scraped_all_brooks_pitch_logs:
                missing_pitch_app_ids = DateScrapeStatus.get_unscraped_pitch_app_ids_for_date(
                    db_session, date_status.game_date
                )
                missing_ids_str = (
                    f"MISSING: {pformat(missing_pitch_app_ids)}"
                    if missing_pitch_app_ids
                    else "All PitchFX logs have been scraped"
                )
            else:
                missing_ids_str = (
                    "Missing IDs cannot be reported until all pitch logs have been scraped."
                )
        print_message(f"\n### STATUS REPORT FOR {game_date_str} ###", fg="bright_cyan", bold=True)
        print_message(date_status.status_report(), wrap=False, fg="bright_cyan")
        if missing_pitchfx:
            print_message(missing_ids_str, wrap=False, fg="bright_cyan")
    return Result.Ok()


def display_summary_report_for_date_range(db_session, start_date, end_date, status_date_range):
    start_str = start_date.strftime(DATE_MONTH_NAME)
    end_str = end_date.strftime(DATE_MONTH_NAME)
    print_message(
        f"\n### STATUS REPORT FOR {start_str} - {end_str} ###", fg="bright_magenta", bold=True
    )
    if not status_date_range:
        print_message("All data has been scraped for all dates in the requested range")
        return Result.Ok()
    for status in status_date_range:
        date_str = status.game_date_str
        status_description = status.scrape_status_description
        print_message(f"{date_str}: {status_description}", wrap=False, fg="bright_magenta")
    return Result.Ok()



import json
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from dateutil import tz

from vigorish.util.list_helpers import as_dict_list


@dataclass
class BrooksPitchFxLog:
    pitchfx_log: Any
    pitch_count_by_inning: Any
    total_pitch_count: str = "0"
    pitcher_name: str = ""
    pitcher_id_mlb: str = "0"
    pitch_app_id: str = ""
    pitcher_team_id_bb: str = ""
    opponent_team_id_bb: str = ""
    bb_game_id: str = ""
    bbref_game_id: str = ""
    game_date_year: str = ""
    game_date_month: str = ""
    game_date_day: str = ""
    game_time_hour: str = ""
    game_time_minute: str = ""
    time_zone_name: str = ""
    pitchfx_url: str = ""

    @property
    def game_date(self):
        return datetime(self.game_date_year, self.game_date_month, self.game_date_day).date()

    @property
    def game_start_time(self):
        if not self.game_time_hour:
            return None
        game_start_time = datetime(
            year=self.game_date.year,
            month=self.game_date.month,
            day=self.game_date.day,
            hour=self.game_time_hour,
            minute=self.game_time_minute,
        )
        return game_start_time.replace(tzinfo=tz.gettz(self.time_zone_name))

    def as_dict(self):
        return dict(
            __brooks_pitchfx_log__=True,
            pitchfx_log=as_dict_list(self.pitchfx_log),
            pitch_count_by_inning=self.pitch_count_by_inning,
            pitcher_name=self.pitcher_name,
            pitcher_id_mlb=int(self.pitcher_id_mlb),
            pitch_app_id=self.pitch_app_id,
            total_pitch_count=int(self.total_pitch_count),
            pitcher_team_id_bb=self.pitcher_team_id_bb,
            opponent_team_id_bb=self.opponent_team_id_bb,
            bb_game_id=self.bb_game_id,
            bbref_game_id=self.bbref_game_id,
            game_date_year=int(self.game_date_year),
            game_date_month=int(self.game_date_month),
            game_date_day=int(self.game_date_day),
            game_time_hour=int(self.game_time_hour),
            game_time_minute=int(self.game_time_minute),
            time_zone_name=self.time_zone_name,
            pitchfx_url=self.pitchfx_url,
        )

    def as_json(self):
        """Convert pitchfx log to JSON."""
        return json.dumps(self.as_dict(), indent=2)



from sqlalchemy import and_, func, join, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, PitchAppScrapeStatus, Season
from vigorish.enums import SeasonType
from vigorish.util.list_helpers import flatten_list2d


class Season_Game_PitchApp_View(Base):
    __table__ = create_view(
        name="season_game_pitch_app_status",
        selectable=select(
            [
                Season.id.label("id"),
                Season.year.label("year"),
                Season.season_type.label("season_type"),
                PitchAppScrapeStatus.scrape_status_date_id.label("date_id"),
                PitchAppScrapeStatus.scrape_status_game_id.label("game_id"),
                PitchAppScrapeStatus.bbref_game_id.label("bbref_game_id"),
                func.count(PitchAppScrapeStatus.id).label("total_pitchfx"),
                func.sum(PitchAppScrapeStatus.scraped_pitchfx).label("total_pitchfx_scraped"),
                func.sum(PitchAppScrapeStatus.no_pitchfx_data).label("total_no_pitchfx_data"),
                func.sum(PitchAppScrapeStatus.combined_pitchfx_bbref_data).label(
                    "total_combined_pitchfx_bbref_data"
                ),
                func.sum(PitchAppScrapeStatus.batters_faced_bbref).label(
                    "total_batters_faced_bbref"
                ),
                func.sum(PitchAppScrapeStatus.batters_faced_pitchfx).label(
                    "total_batters_faced_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.duplicate_guid_removed_count).label(
                    "total_duplicate_pitchfx_removed_count"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_pitch_log).label(
                    "total_pitch_count_pitch_log"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_bbref).label("total_pitch_count_bbref"),
                func.sum(PitchAppScrapeStatus.pitch_count_pitchfx).label(
                    "total_pitch_count_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.pitch_count_pitchfx_audited).label(
                    "total_pitch_count_pitchfx_audited"
                ),
                func.sum(PitchAppScrapeStatus.patched_pitchfx_count).label(
                    "total_patched_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_patched_pitchfx).label(
                    "total_at_bats_patched_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.missing_pitchfx_count).label(
                    "total_missing_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_missing_pitchfx).label(
                    "total_at_bats_missing_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.extra_pitchfx_count).label(
                    "total_extra_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_extra_pitchfx).label(
                    "total_at_bats_extra_pitchfx"
                ),
                func.sum(PitchAppScrapeStatus.extra_pitchfx_removed_count).label(
                    "total_extra_pitchfx_removed_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_extra_pitchfx_removed).label(
                    "total_at_bats_extra_pitchfx_removed"
                ),
                func.sum(PitchAppScrapeStatus.pitchfx_error).label("total_pitchfx_error"),
                func.sum(PitchAppScrapeStatus.total_at_bats_pitchfx_error).label(
                    "total_at_bats_pitchfx_error"
                ),
                func.sum(PitchAppScrapeStatus.invalid_pitchfx).label("total_invalid_pitchfx"),
                func.sum(PitchAppScrapeStatus.invalid_pitchfx_count).label(
                    "total_invalid_pitchfx_count"
                ),
                func.sum(PitchAppScrapeStatus.total_at_bats_invalid_pitchfx).label(
                    "total_at_bats_invalid_pitchfx"
                ),
            ]
        )
        .select_from(
            join(
                Season,
                PitchAppScrapeStatus,
                Season.id == PitchAppScrapeStatus.season_id,
                isouter=True,
            )
        )
        .group_by(PitchAppScrapeStatus.scrape_status_game_id),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_all_bbref_game_ids_no_pitchfx_data_for_any_pitch_apps(
        cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.year == year,
                Season_Game_PitchApp_View.total_pitchfx
                == Season_Game_PitchApp_View.total_pitchfx_scraped,
                Season_Game_PitchApp_View.total_no_pitchfx_data > 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_eligible_for_audit(
        cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.year == year,
                Season_Game_PitchApp_View.total_pitchfx
                == Season_Game_PitchApp_View.total_pitchfx_scraped,
                Season_Game_PitchApp_View.total_pitchfx
                != Season_Game_PitchApp_View.total_combined_pitchfx_bbref_data,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_all_pitchfx_logs_are_valid(
        cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.year == year,
                Season_Game_PitchApp_View.total_pitchfx
                == Season_Game_PitchApp_View.total_combined_pitchfx_bbref_data,
                Season_Game_PitchApp_View.total_pitchfx_error == 0,
                Season_Game_PitchApp_View.total_invalid_pitchfx == 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_pitchfx_error(
        cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.year == year,
                Season_Game_PitchApp_View.total_pitchfx_error > 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_invalid_pitchfx(
        cls, db_engine, year, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.year == year,
                Season_Game_PitchApp_View.total_invalid_pitchfx > 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])

    @classmethod
    def get_all_bbref_game_ids_combined_no_missing_pfx(
        cls, db_engine, season_type=SeasonType.REGULAR_SEASON
    ):
        s = select([Season_Game_PitchApp_View.bbref_game_id]).where(
            and_(
                Season_Game_PitchApp_View.total_pitchfx
                == Season_Game_PitchApp_View.total_combined_pitchfx_bbref_data,
                Season_Game_PitchApp_View.total_missing_pitchfx_count == 0,
                Season_Game_PitchApp_View.total_pitchfx_error == 0,
                Season_Game_PitchApp_View.total_invalid_pitchfx == 0,
            )
        )
        results = db_engine.execute(s).fetchall()
        return flatten_list2d([d.values() for d in [dict(row) for row in results]])
